"""Type stubs for SLEPc MFN module."""

from enum import IntEnum, StrEnum
from typing import Any, Callable

from petsc4py.PETSc import Comm, Mat, Vec, Viewer

from .BV import BV
from .FN import FN

class MFNType(StrEnum):
    """MFN type - action of a matrix function on a vector."""

    KRYLOV = ...
    """Krylov method."""
    EXPOKIT = ...
    """Expokit method."""

class MFNConvergedReason(IntEnum):
    """MFN convergence reasons."""

    CONVERGED_TOL = ...
    """Converged to requested tolerance."""
    CONVERGED_ITS = ...
    """Converged within iterations."""
    DIVERGED_ITS = ...
    """Maximum number of iterations exceeded."""
    DIVERGED_BREAKDOWN = ...
    """Solver failed due to breakdown."""
    CONVERGED_ITERATING = ...
    """Iteration not finished yet."""
    ITERATING = ...
    """Iteration not finished yet (alias)."""

class MFN:
    """Matrix Function solver - computes f(A)b."""

    Type: type[MFNType]
    ConvergedReason: type[MFNConvergedReason]

    def __init__(self) -> None: ...

    # --- Basic operations ---

    def view(self, viewer: Viewer | None = None) -> None:
        """
        Print the MFN data structure.

        Parameters
        ----------
        viewer
            Viewer to print the MFN object.
        """
        ...

    def destroy(self) -> MFN:
        """Destroy the MFN object."""
        ...

    def reset(self) -> None:
        """Reset the MFN object to its initial state."""
        ...

    def create(self, comm: Comm | None = None) -> MFN:
        """
        Create the MFN object.

        Parameters
        ----------
        comm
            MPI communicator.

        Returns
        -------
        MFN
            The created MFN object.
        """
        ...

    def setType(self, mfn_type: MFNType | str) -> None:
        """
        Set the type of the MFN object.

        Parameters
        ----------
        mfn_type
            The type of MFN solver to use.
        """
        ...

    def getType(self) -> str:
        """
        Get the type of the MFN object.

        Returns
        -------
        str
            The type of MFN solver.
        """
        ...

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for all MFN options in the database.

        Returns
        -------
        str
            The options prefix string.
        """
        ...

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for all MFN options in the database.

        Parameters
        ----------
        prefix
            The prefix string.
        """
        ...

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for all MFN options in the database.

        Parameters
        ----------
        prefix
            The prefix string to append.
        """
        ...

    def setFromOptions(self) -> None:
        """Set MFN options from the options database."""
        ...

    # --- Tolerances ---

    def getTolerances(self) -> tuple[float, int]:
        """
        Get the tolerance and maximum iteration count.

        Returns
        -------
        tol : float
            The convergence tolerance.
        max_it : int
            Maximum number of iterations.
        """
        ...

    def setTolerances(
        self, tol: float | None = None, max_it: int | None = None
    ) -> None:
        """
        Set the tolerance and maximum iteration count.

        Parameters
        ----------
        tol
            The convergence tolerance.
        max_it
            Maximum number of iterations.
        """
        ...

    # --- Dimensions ---

    def getDimensions(self) -> int:
        """
        Get the dimension of the subspace used by the solver.

        Returns
        -------
        int
            Maximum dimension of the subspace.
        """
        ...

    def setDimensions(self, ncv: int) -> None:
        """
        Set the dimension of the subspace to be used by the solver.

        Parameters
        ----------
        ncv
            Maximum dimension of the subspace.
        """
        ...

    # --- Associated objects ---

    def getFN(self) -> FN:
        """
        Get the math function object associated with the MFN object.

        Returns
        -------
        FN
            The math function context.
        """
        ...

    def setFN(self, fn: FN) -> None:
        """
        Set a math function object associated with the MFN object.

        Parameters
        ----------
        fn
            The math function context.
        """
        ...

    def getBV(self) -> BV:
        """
        Get the basis vector object associated with the MFN object.

        Returns
        -------
        BV
            The basis vectors context.
        """
        ...

    def setBV(self, bv: BV) -> None:
        """
        Set a basis vector object associated with the MFN object.

        Parameters
        ----------
        bv
            The basis vectors context.
        """
        ...

    def getOperator(self) -> Mat:
        """
        Get the matrix associated with the MFN object.

        Returns
        -------
        Mat
            The matrix for which the matrix function is to be computed.
        """
        ...

    def setOperator(self, A: Mat) -> None:
        """
        Set the matrix associated with the MFN object.

        Parameters
        ----------
        A
            The problem matrix.
        """
        ...

    # --- Monitor ---

    def setMonitor(
        self,
        monitor: Callable[..., None] | None,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Append a monitor function to the list of monitors.

        Parameters
        ----------
        monitor
            A callable to monitor the solver.
        args
            Positional arguments for the callable.
        kargs
            Keyword arguments for the callable.
        """
        ...

    def getMonitor(self) -> list[tuple[Callable[..., None], tuple, dict]] | None:
        """
        Get the list of monitor functions.

        Returns
        -------
        list or None
            The list of monitor functions.
        """
        ...

    def cancelMonitor(self) -> None:
        """Clear all monitors for the MFN object."""
        ...

    # --- Solve ---

    def setUp(self) -> None:
        """Set up all the necessary internal data structures."""
        ...

    def solve(self, b: Vec, x: Vec) -> None:
        """
        Solve the matrix function problem: x = f(A)b.

        Parameters
        ----------
        b
            The right hand side vector.
        x
            The solution.
        """
        ...

    def solveTranspose(self, b: Vec, x: Vec) -> None:
        """
        Solve the transpose matrix function problem: x = f(A^T)b.

        Parameters
        ----------
        b
            The right hand side vector.
        x
            The solution.
        """
        ...

    def getIterationNumber(self) -> int:
        """
        Get the current iteration number.

        Returns
        -------
        int
            Iteration number.
        """
        ...

    def getConvergedReason(self) -> MFNConvergedReason:
        """
        Get the reason why the solve iteration was stopped.

        Returns
        -------
        MFNConvergedReason
            Negative value indicates diverged, positive value converged.
        """
        ...

    def setErrorIfNotConverged(self, flg: bool = True) -> None:
        """
        Set solve() to generate an error if the solver does not converge.

        Parameters
        ----------
        flg
            True indicates you want the error generated.
        """
        ...

    def getErrorIfNotConverged(self) -> bool:
        """
        Get if solve() generates an error if the solver does not converge.

        Returns
        -------
        bool
            True indicates the error will be generated.
        """
        ...

    # --- Properties ---

    @property
    def tol(self) -> float:
        """The tolerance used by the MFN convergence tests."""
        ...

    @tol.setter
    def tol(self, value: float) -> None: ...
    @property
    def max_it(self) -> int:
        """The maximum iteration count used by the MFN convergence tests."""
        ...

    @max_it.setter
    def max_it(self, value: int) -> None: ...
    @property
    def fn(self) -> FN:
        """The math function (FN) object associated with the MFN object."""
        ...

    @fn.setter
    def fn(self, value: FN) -> None: ...
    @property
    def bv(self) -> BV:
        """The basis vectors (BV) object associated with the MFN object."""
        ...

    @bv.setter
    def bv(self, value: BV) -> None: ...



"""Type stubs for SLEPc LME module."""

from enum import IntEnum, StrEnum
from typing import Any, Self

from petsc4py.PETSc import Comm, Mat, Viewer
from slepc4py.typing import LMEMonitorFunction

from .BV import BV

class LMEType(StrEnum):
    """LME type.

    - `KRYLOV`:  Restarted Krylov solver.
    """

    KRYLOV = ...
    """Restarted Krylov solver."""

class LMEConvergedReason(IntEnum):
    """LME convergence reasons.

    - `CONVERGED_TOL`:       All eigenpairs converged to requested tolerance.
    - `DIVERGED_ITS`:        Maximum number of iterations exceeded.
    - `DIVERGED_BREAKDOWN`:  Solver failed due to breakdown.
    - `CONVERGED_ITERATING`: Iteration not finished yet.
    """

    CONVERGED_TOL = ...
    """All eigenpairs converged to requested tolerance."""
    DIVERGED_ITS = ...
    """Maximum number of iterations exceeded."""
    DIVERGED_BREAKDOWN = ...
    """Solver failed due to breakdown."""
    CONVERGED_ITERATING = ...
    """Iteration not finished yet."""
    ITERATING = ...
    """Iteration not finished yet (alias)."""

class LMEProblemType(IntEnum):
    """LME problem type.

    - `LYAPUNOV`:      Continuous-time Lyapunov.
    - `SYLVESTER`:     Continuous-time Sylvester.
    - `GEN_LYAPUNOV`:  Generalized Lyapunov.
    - `GEN_SYLVESTER`: Generalized Sylvester.
    - `DT_LYAPUNOV`:   Discrete-time Lyapunov.
    - `STEIN`:         Stein.
    """

    LYAPUNOV = ...
    """Continuous-time Lyapunov."""
    SYLVESTER = ...
    """Continuous-time Sylvester."""
    GEN_LYAPUNOV = ...
    """Generalized Lyapunov."""
    GEN_SYLVESTER = ...
    """Generalized Sylvester."""
    DT_LYAPUNOV = ...
    """Discrete-time Lyapunov."""
    STEIN = ...
    """Stein."""

class LME:
    """Linear matrix equation solver.

    LME is used to solve linear matrix equations such as Lyapunov
    and Sylvester equations.

    See Also
    --------
    slepc.LME
    """

    Type = LMEType
    ProblemType = LMEProblemType
    ConvergedReason = LMEConvergedReason

    # --- View and Lifecycle ---

    def view(self, viewer: Viewer | None = None) -> None:
        """Print the LME data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.

        See Also
        --------
        slepc.LMEView
        """
        ...

    def destroy(self) -> Self:
        """Destroy the LME object.

        Collective.

        See Also
        --------
        slepc.LMEDestroy
        """
        ...

    def reset(self) -> None:
        """Reset the LME object.

        Collective.

        See Also
        --------
        slepc.LMEReset
        """
        ...

    def create(self, comm: Comm | None = None) -> Self:
        """Create the LME object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator. If not provided, it defaults to all processes.

        See Also
        --------
        slepc.LMECreate
        """
        ...

    # --- Type and Problem Type ---

    def setType(self, lme_type: LMEType | str) -> None:
        """Set the particular solver to be used in the LME object.

        Logically collective.

        Parameters
        ----------
        lme_type
            The solver to be used.

        See Also
        --------
        slepc.LMESetType
        """
        ...

    def getType(self) -> str:
        """Get the LME type of this object.

        Not collective.

        Returns
        -------
        str
            The solver currently being used.

        See Also
        --------
        slepc.LMEGetType
        """
        ...

    def setProblemType(self, lme_problem_type: LMEProblemType) -> None:
        """Set the LME problem type of this object.

        Logically collective.

        Parameters
        ----------
        lme_problem_type
            The problem type to be used.

        See Also
        --------
        slepc.LMESetProblemType
        """
        ...

    def getProblemType(self) -> LMEProblemType:
        """Get the LME problem type of this object.

        Not collective.

        Returns
        -------
        LMEProblemType
            The problem type currently being used.

        See Also
        --------
        slepc.LMEGetProblemType
        """
        ...

    # --- Coefficient Matrices ---

    def setCoefficients(
        self,
        A: Mat,
        B: Mat | None = None,
        D: Mat | None = None,
        E: Mat | None = None,
    ) -> None:
        """Set the coefficient matrices.

        Collective.

        Set the coefficient matrices that define the linear matrix equation
        to be solved.

        Parameters
        ----------
        A
            First coefficient matrix.
        B
            Second coefficient matrix, optional.
        D
            Third coefficient matrix, optional.
        E
            Fourth coefficient matrix, optional.

        See Also
        --------
        slepc.LMESetCoefficients
        """
        ...

    def getCoefficients(self) -> tuple[Mat, Mat | None, Mat | None, Mat | None]:
        """Get the coefficient matrices of the matrix equation.

        Collective.

        Returns
        -------
        A : Mat
            First coefficient matrix.
        B : Mat | None
            Second coefficient matrix, if available.
        D : Mat | None
            Third coefficient matrix, if available.
        E : Mat | None
            Fourth coefficient matrix, if available.

        See Also
        --------
        slepc.LMEGetCoefficients
        """
        ...

    # --- RHS and Solution ---

    def setRHS(self, C: Mat) -> None:
        """Set the right-hand side of the matrix equation.

        Collective.

        Set the right-hand side of the matrix equation, as a low-rank
        matrix.

        Parameters
        ----------
        C
            The right-hand side matrix.

        See Also
        --------
        slepc.LMESetRHS
        """
        ...

    def getRHS(self) -> Mat:
        """Get the right-hand side of the matrix equation.

        Collective.

        Returns
        -------
        Mat
            The low-rank matrix.

        See Also
        --------
        slepc.LMEGetRHS
        """
        ...

    def setSolution(self, X: Mat) -> None:
        """Set the placeholder for the solution of the matrix equation.

        Collective.

        Set the placeholder for the solution of the matrix
        equation, as a low-rank matrix.

        Parameters
        ----------
        X
            The solution matrix.

        See Also
        --------
        slepc.LMESetSolution
        """
        ...

    def getSolution(self) -> Mat:
        """Get the solution of the matrix equation.

        Collective.

        Returns
        -------
        Mat
            The low-rank matrix.

        See Also
        --------
        slepc.LMEGetSolution
        """
        ...

    # --- Error and Options ---

    def getErrorEstimate(self) -> float:
        """Get the error estimate obtained during solve.

        Not collective.

        Returns
        -------
        float
            The error estimate.

        See Also
        --------
        slepc.LMEGetErrorEstimate
        """
        ...

    def computeError(self) -> float:
        """Compute the error associated with the last equation solved.

        Collective.

        Computes the error (based on the residual norm) associated with the
        last equation solved.

        Returns
        -------
        float
            The error.

        See Also
        --------
        slepc.LMEComputeError
        """
        ...

    def getOptionsPrefix(self) -> str:
        """Get the prefix used for searching for all LME options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this LME object.

        See Also
        --------
        slepc.LMEGetOptionsPrefix
        """
        ...

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """Set the prefix used for searching for all LME options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all LME option requests.

        See Also
        --------
        slepc.LMESetOptionsPrefix
        """
        ...

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """Append to the prefix used for searching in the database.

        Logically collective.

        Append to the prefix used for searching for all LME options in the
        database.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all LME option requests.

        See Also
        --------
        slepc.LMEAppendOptionsPrefix
        """
        ...

    def setFromOptions(self) -> None:
        """Set LME options from the options database.

        Collective.

        Sets LME options from the options database. This routine must
        be called before `setUp()` if the user is to be allowed to set
        the solver type.

        See Also
        --------
        slepc.LMESetFromOptions
        """
        ...

    # --- Tolerances and Dimensions ---

    def getTolerances(self) -> tuple[float, int]:
        """Get the tolerance and maximum iteration count.

        Not collective.

        Get the tolerance and maximum iteration count used by the
        default LME convergence tests.

        Returns
        -------
        tol : float
            The convergence tolerance.
        max_it : int
            The maximum number of iterations.

        See Also
        --------
        slepc.LMEGetTolerances
        """
        ...

    def setTolerances(
        self,
        tol: float | None = None,
        max_it: int | None = None,
    ) -> None:
        """Set the tolerance and maximum iteration count.

        Logically collective.

        Set the tolerance and maximum iteration count used by the
        default LME convergence tests.

        Parameters
        ----------
        tol
            The convergence tolerance.
        max_it
            The maximum number of iterations.

        See Also
        --------
        slepc.LMESetTolerances
        """
        ...

    def getDimensions(self) -> int:
        """Get the dimension of the subspace used by the solver.

        Not collective.

        Returns
        -------
        int
            Maximum dimension of the subspace to be used by the solver.

        See Also
        --------
        slepc.LMEGetDimensions
        """
        ...

    def setDimensions(self, ncv: int) -> None:
        """Set the dimension of the subspace to be used by the solver.

        Logically collective.

        Parameters
        ----------
        ncv
            Maximum dimension of the subspace to be used by the solver.

        See Also
        --------
        slepc.LMESetDimensions
        """
        ...

    # --- BV ---

    def getBV(self) -> BV:
        """Get the basis vector object associated to the LME object.

        Not collective.

        Returns
        -------
        BV
            The basis vectors context.

        See Also
        --------
        slepc.LMEGetBV
        """
        ...

    def setBV(self, bv: BV) -> None:
        """Set a basis vector object to the LME object.

        Collective.

        Parameters
        ----------
        bv
            The basis vectors context.

        See Also
        --------
        slepc.LMESetBV
        """
        ...

    # --- Monitor ---

    def setMonitor(
        self,
        monitor: LMEMonitorFunction | None,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None,
    ) -> None:
        """Append a monitor function to the list of monitors.

        Logically collective.

        Parameters
        ----------
        monitor
            The monitor function.
        args
            Positional arguments for the monitor.
        kargs
            Keyword arguments for the monitor.
        """
        ...

    def getMonitor(
        self,
    ) -> list[tuple[LMEMonitorFunction, tuple[Any, ...], dict[str, Any]]] | None:
        """Get the list of monitor functions.

        Returns
        -------
        list | None
            The list of monitor functions with their arguments.
        """
        ...

    def cancelMonitor(self) -> None:
        """Clear all monitors for an `LME` object.

        See Also
        --------
        slepc.LMEMonitorCancel
        """
        ...

    # --- Solve ---

    def setUp(self) -> None:
        """Set up all the internal necessary data structures.

        Collective.

        Set up all the internal data structures necessary for the
        execution of the solver.

        See Also
        --------
        slepc.LMESetUp
        """
        ...

    def solve(self) -> None:
        """Solve the linear matrix equation.

        Collective.

        See Also
        --------
        slepc.LMESolve
        """
        ...

    def getIterationNumber(self) -> int:
        """Get the current iteration number.

        Not collective.

        If the call to `solve()` is complete, then it returns the number of
        iterations carried out by the solution method.

        Returns
        -------
        int
            Iteration number.

        See Also
        --------
        slepc.LMEGetIterationNumber
        """
        ...

    def getConvergedReason(self) -> LMEConvergedReason:
        """Get the reason why the `solve()` iteration was stopped.

        Not collective.

        Returns
        -------
        LMEConvergedReason
            Negative value indicates diverged, positive value converged.

        See Also
        --------
        slepc.LMEGetConvergedReason
        """
        ...

    def setErrorIfNotConverged(self, flg: bool = True) -> None:
        """Set `solve()` to generate an error if the solver has not converged.

        Logically collective.

        Parameters
        ----------
        flg
            True indicates you want the error generated.

        See Also
        --------
        slepc.LMESetErrorIfNotConverged
        """
        ...

    def getErrorIfNotConverged(self) -> bool:
        """Get if `solve()` generates an error if the solver does not converge.

        Not collective.

        Get a flag indicating whether `solve()` will generate an error if the
        solver does not converge.

        Returns
        -------
        bool
            True indicates you want the error generated.

        See Also
        --------
        slepc.LMEGetErrorIfNotConverged
        """
        ...

    # --- Properties ---

    @property
    def tol(self) -> float:
        """The tolerance value used by the LME convergence tests."""
        ...

    @tol.setter
    def tol(self, value: float) -> None: ...
    @property
    def max_it(self) -> int:
        """The maximum iteration count used by the LME convergence tests."""
        ...

    @max_it.setter
    def max_it(self, value: int) -> None: ...
    @property
    def bv(self) -> BV:
        """The basis vectors (BV) object associated to the LME object."""
        ...

    @bv.setter
    def bv(self, value: BV) -> None: ...



# -----------------------------------------------------------------------------

class FNType(object):
    """FN type."""
    COMBINE  = S_(FNCOMBINE)
    RATIONAL = S_(FNRATIONAL)
    EXP      = S_(FNEXP)
    LOG      = S_(FNLOG)
    PHI      = S_(FNPHI)
    SQRT     = S_(FNSQRT)
    INVSQRT  = S_(FNINVSQRT)

class FNCombineType(object):
    """
    FN type of combination of child functions.

    - `ADD`:       Addition         f(x) = f1(x)+f2(x)
    - `MULTIPLY`:  Multiplication   f(x) = f1(x)*f2(x)
    - `DIVIDE`:    Division         f(x) = f1(x)/f2(x)
    - `COMPOSE`:   Composition      f(x) = f2(f1(x))
    """
    ADD      = FN_COMBINE_ADD
    MULTIPLY = FN_COMBINE_MULTIPLY
    DIVIDE   = FN_COMBINE_DIVIDE
    COMPOSE  = FN_COMBINE_COMPOSE

class FNParallelType(object):
    """
    FN parallel types.

    - `REDUNDANT`:    Every process performs the computation redundantly.
    - `SYNCHRONIZED`: The first process sends the result to the rest.
    """
    REDUNDANT    = FN_PARALLEL_REDUNDANT
    SYNCHRONIZED = FN_PARALLEL_SYNCHRONIZED

# -----------------------------------------------------------------------------

cdef class FN(Object):

    """FN."""

    Type         = FNType
    CombineType  = FNCombineType
    ParallelType = FNParallelType

    def __cinit__(self):
        self.obj = <PetscObject*> &self.fn
        self.fn = NULL

    # unary operations

    def __pos__(self):
        return fn_pos(self)

    def __neg__(self):
        return fn_neg(self)

    # inplace binary operations

    def __iadd__(self, other):
        return fn_iadd(self, other)

    def __isub__(self, other):
        return fn_isub(self, other)

    def __imul__(self, other):
        return fn_imul(self, other)

    def __idiv__(self, other):
        return fn_idiv(self, other)

    def __itruediv__(self, other):
        return fn_idiv(self, other)

    # binary operations

    def __add__(self, other):
        return fn_add(self, other)

    def __radd__(self, other):
        return fn_radd(self, other)

    def __sub__(self, other):
        return fn_sub(self, other)

    def __rsub__(self, other):
        return fn_rsub(self, other)

    def __mul__(self, other):
        return fn_mul(self, other)

    def __rmul__(self, other):
        return fn_rmul(self, other)

    def __div__(self, other):
        return fn_div(self, other)

    def __rdiv__(self, other):
        return fn_rdiv(self, other)

    def __truediv__(self, other):
        return fn_div(self, other)

    def __rtruediv__(self, other):
        return fn_rdiv(self, other)

    def __matmul__(self, other):
        return fn_matmul(self, other)

    def __call__(self, arg):
        if isinstance(arg, Mat):
            return self.evaluateFunctionMat(arg)
        else:
            return self.evaluateFunction(arg)

    #

    def view(self, Viewer viewer=None) -> None:
        """
        Print the FN data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.
        """
        cdef PetscViewer vwr = def_Viewer(viewer)
        CHKERR( FNView(self.fn, vwr) )

    def destroy(self) -> Self:
        """
        Destroy the FN object.

        Collective.
        """
        CHKERR( FNDestroy(&self.fn) )
        self.fn = NULL
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """
        Create the FN object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to all processes.
        """
        cdef MPI_Comm ccomm = def_Comm(comm, SLEPC_COMM_DEFAULT())
        cdef SlepcFN newfn = NULL
        CHKERR( FNCreate(ccomm, &newfn) )
        CHKERR( SlepcCLEAR(self.obj) ); self.fn = newfn
        return self

    def setType(self, fn_type: Type | str) -> None:
        """
        Set the type for the FN object.

        Logically collective.

        Parameters
        ----------
        fn_type
            The inner product type to be used.
        """
        cdef SlepcFNType cval = NULL
        fn_type = str2bytes(fn_type, &cval)
        CHKERR( FNSetType(self.fn, cval) )

    def getType(self) -> str:
        """
        Get the FN type of this object.

        Not collective.

        Returns
        -------
        str
            The inner product type currently being used.
        """
        cdef SlepcFNType fn_type = NULL
        CHKERR( FNGetType(self.fn, &fn_type) )
        return bytes2str(fn_type)

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for searching for all FN options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all FN option requests.

        Notes
        -----
        A hyphen (``-``) must NOT be given at the beginning of the
        prefix name.  The first character of all runtime options is
        AUTOMATICALLY the hyphen.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( FNSetOptionsPrefix(self.fn, cval) )

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for searching for all FN options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all FN option requests.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( FNAppendOptionsPrefix(self.fn, cval) )

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for searching for all FN options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this FN object.
        """
        cdef const char *prefix = NULL
        CHKERR( FNGetOptionsPrefix(self.fn, &prefix) )
        return bytes2str(prefix)

    def setFromOptions(self) -> None:
        """
        Set FN options from the options database.

        Collective.

        Notes
        -----
        To see all options, run your program with the ``-help``
        option.
        """
        CHKERR( FNSetFromOptions(self.fn) )

    def duplicate(self, comm: Comm | None = None) -> FN:
        """
        Duplicate the FN object copying all parameters.

        Collective.

        Duplicate the FN object copying all parameters, possibly with a
        different communicator.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to the
            object's communicator.
        """
        cdef MPI_Comm ccomm = def_Comm(comm, PetscObjectComm(<PetscObject>self.fn))
        cdef FN fn = type(self)()
        CHKERR( FNDuplicate(self.fn, ccomm, &fn.fn) )
        return fn

    #

    def evaluateFunction(self, x: Scalar) -> Scalar:
        """
        Compute the value of the function f(x) for a given x.

        Not collective.

        Parameters
        ----------
        x
            Value where the function must be evaluated.

        Returns
        -------
        Scalar
            The result of f(x).
        """
        cdef PetscScalar sval = 0
        cdef PetscScalar sarg = asScalar(x)
        CHKERR( FNEvaluateFunction(self.fn, sarg, &sval) )
        return toScalar(sval)

    def evaluateDerivative(self, x: Scalar) -> Scalar:
        """
        Compute the value of the derivative :math:`f'(x)` for a given x.

        Not collective.

        Parameters
        ----------
        x
            Value where the derivative must be evaluated.

        Returns
        -------
        Scalar
            The result of :math:`f'(x)`.
        """
        cdef PetscScalar sval = 0
        cdef PetscScalar sarg = asScalar(x)
        CHKERR( FNEvaluateDerivative(self.fn, sarg, &sval) )
        return toScalar(sval)

    def evaluateFunctionMat(self, Mat A: petsc4py.PETSc.Mat, Mat B: petsc4py.PETSc.Mat | None = None) -> petsc4py.PETSc.Mat:
        """
        Compute the value of the function :math:`f(A)` for a given matrix A.

        Logically collective.

        Parameters
        ----------
        A
            Matrix on which the function must be evaluated.
        B
            Placeholder for the result.

        Returns
        -------
        petsc4py.PETSc.Mat
            The result of :math:`f(A)`.
        """
        if B is None: B = A.duplicate()
        CHKERR( FNEvaluateFunctionMat(self.fn, A.mat, B.mat) )
        return B

    def evaluateFunctionMatVec(self, Mat A: petsc4py.PETSc.Mat, Vec v: petsc4py.PETSc.Vec | None = None) -> petsc4py.PETSc.Vec:
        """
        Compute the first column of the matrix f(A) for a given matrix A.

        Logically collective.

        Parameters
        ----------
        A
            Matrix on which the function must be evaluated.

        Returns
        -------
        petsc4py.PETSc.Vec
            The first column of the result f(A).
        """
        if v is None: v = A.createVecs('left')
        CHKERR( FNEvaluateFunctionMatVec(self.fn, A.mat, v.vec) )
        return v

    def setScale(self, alpha: Scalar | None = None, beta: Scalar | None = None) -> None:
        """
        Set the scaling parameters that define the matematical function.

        Logically collective.

        Parameters
        ----------
        alpha
            Inner scaling (argument), default is 1.0.
        beta
            Outer scaling (result), default is 1.0.
        """
        cdef PetscScalar aval = 1.0
        cdef PetscScalar bval = 1.0
        if alpha is not None: aval = asScalar(alpha)
        if beta  is not None: bval = asScalar(beta)
        CHKERR( FNSetScale(self.fn, aval, bval) )

    def getScale(self) -> tuple[Scalar, Scalar]:
        """
        Get the scaling parameters that define the matematical function.

        Not collective.

        Returns
        -------
        alpha: Scalar
            Inner scaling (argument).
        beta: Scalar
            Outer scaling (result).
        """
        cdef PetscScalar aval = 0, bval = 0
        CHKERR( FNGetScale(self.fn, &aval, &bval) )
        return (toScalar(aval), toScalar(bval))

    def setMethod(self, meth: int) -> None:
        """
        Set the method to be used to evaluate functions of matrices.

        Logically collective.

        Parameters
        ----------
        meth
            An index identifying the method.

        Notes
        -----
        In some `FN` types there are more than one algorithms available
        for computing matrix functions. In that case, this function allows
        choosing the wanted method.

        If ``meth`` is currently set to 0 and the input argument of
        `FN.evaluateFunctionMat()` is a symmetric/Hermitian matrix, then
        the computation is done via the eigendecomposition, rather than
        with the general algorithm.
        """
        cdef PetscInt val = asInt(meth)
        CHKERR( FNSetMethod(self.fn, val) )

    def getMethod(self) -> int:
        """
        Get the method currently used for matrix functions.

        Not collective.

        Returns
        -------
        int
            An index identifying the method.
        """
        cdef PetscInt val = 0
        CHKERR( FNGetMethod(self.fn, &val) )
        return toInt(val)

    def setParallel(self, pmode: ParallelType) -> None:
        """
        Set the mode of operation in parallel runs.

        Logically collective.

        Parameters
        ----------
        pmode
            The parallel mode.
        """
        cdef SlepcFNParallelType val = pmode
        CHKERR( FNSetParallel(self.fn, val) )

    def getParallel(self) -> ParallelType:
        """
        Get the mode of operation in parallel runs.

        Not collective.

        Returns
        -------
        ParallelType
            The parallel mode.
        """
        cdef SlepcFNParallelType val = FN_PARALLEL_REDUNDANT
        CHKERR( FNGetParallel(self.fn, &val) )
        return val

    #

    def setRationalNumerator(self, alpha: Sequence[Scalar]) -> None:
        """
        Set the coefficients of the numerator of the rational function.

        Logically collective.

        Parameters
        ----------
        alpha
            Coefficients.
        """
        cdef PetscInt na = 0
        cdef PetscScalar *a = NULL
        cdef object tmp1 = iarray_s(alpha, &na, &a)
        CHKERR( FNRationalSetNumerator(self.fn, na, a) )

    def getRationalNumerator(self) -> ArrayScalar:
        """
        Get the coefficients of the numerator of the rational function.

        Not collective.

        Returns
        -------
        ArrayScalar
            Coefficients.
        """
        cdef PetscInt np = 0
        cdef PetscScalar *coeff = NULL
        CHKERR( FNRationalGetNumerator(self.fn, &np, &coeff) )
        cdef object ocoeff = None
        try:
            ocoeff = array_s(np, coeff)
        finally:
            CHKERR( PetscFree(coeff) )
        return ocoeff

    def setRationalDenominator(self, alpha: Sequence[Scalar]) -> None:
        """
        Set the coefficients of the denominator of the rational function.

        Logically collective.

        Parameters
        ----------
        alpha
            Coefficients.
        """
        cdef PetscInt na = 0
        cdef PetscScalar *a = NULL
        cdef object tmp1 = iarray_s(alpha, &na, &a)
        CHKERR( FNRationalSetDenominator(self.fn, na, a) )

    def getRationalDenominator(self) -> ArrayScalar:
        """
        Get the coefficients of the denominator of the rational function.

        Not collective.

        Returns
        -------
        ArrayScalar
            Coefficients.
        """
        cdef PetscInt np = 0
        cdef PetscScalar *coeff = NULL
        CHKERR( FNRationalGetDenominator(self.fn, &np, &coeff) )
        cdef object ocoeff = None
        try:
            ocoeff = array_s(np, coeff)
        finally:
            CHKERR( PetscFree(coeff) )
        return ocoeff

    def setCombineChildren(self, comb: CombineType, FN f1, FN f2) -> None:
        """
        Set the two child functions that constitute this combined function.

        Logically collective.

        Set the two child functions that constitute this combined function,
        and the way they must be combined.

        Parameters
        ----------
        comb
            How to combine the functions (addition, multiplication, division,
            composition).
        f1
            First function.
        f2
            Second function.
        """
        cdef SlepcFNCombineType val = comb
        CHKERR( FNCombineSetChildren(self.fn, val, f1.fn, f2.fn) )

    def getCombineChildren(self) -> tuple[CombineType, FN, FN]:
        """
        Get the two child functions that constitute this combined function.

        Not collective.

        Get the two child functions that constitute this combined
        function, and the way they must be combined.

        Returns
        -------
        comb: CombineType
            How to combine the functions (addition, multiplication, division,
            composition).
        f1: FN
            First function.
        f2: FN
            Second function.
        """
        cdef SlepcFNCombineType comb
        cdef FN f1 = FN()
        cdef FN f2 = FN()
        CHKERR( FNCombineGetChildren(self.fn, &comb, &f1.fn, &f2.fn) )
        CHKERR( PetscINCREF(f1.obj) )
        CHKERR( PetscINCREF(f2.obj) )
        return (comb, f1, f2)

    def setPhiIndex(self, k: int) -> None:
        """
        Set the index of the phi-function.

        Logically collective.

        Parameters
        ----------
        k
            The index.
        """
        cdef PetscInt val = asInt(k)
        CHKERR( FNPhiSetIndex(self.fn, val) )

    def getPhiIndex(self) -> int:
        """
        Get the index of the phi-function.

        Not collective.

        Returns
        -------
        int
            The index.
        """
        cdef PetscInt val = 0
        CHKERR( FNPhiGetIndex(self.fn, &val) )
        return toInt(val)

    #

    property method:
        """The method to be used to evaluate functions of matrices."""
        def __get__(self) -> int:
            return self.getMethod()
        def __set__(self, value):
            self.setMethod(value)

    property parallel:
        """The mode of operation in parallel runs."""
        def __get__(self) -> FNParallelType:
            return self.getParallel()
        def __set__(self, value):
            self.setParallel(value)

# -----------------------------------------------------------------------------

del FNType
del FNCombineType
del FNParallelType

# -----------------------------------------------------------------------------



cdef extern from * nogil:

    ctypedef char* SlepcFNType "const char*"
    SlepcFNType FNCOMBINE
    SlepcFNType FNRATIONAL
    SlepcFNType FNEXP
    SlepcFNType FNLOG
    SlepcFNType FNPHI
    SlepcFNType FNSQRT
    SlepcFNType FNINVSQRT

    ctypedef enum SlepcFNCombineType "FNCombineType":
        FN_COMBINE_ADD
        FN_COMBINE_MULTIPLY
        FN_COMBINE_DIVIDE
        FN_COMBINE_COMPOSE

    ctypedef enum SlepcFNParallelType "FNParallelType":
        FN_PARALLEL_REDUNDANT
        FN_PARALLEL_SYNCHRONIZED

    PetscErrorCode FNCreate(MPI_Comm,SlepcFN*)
    PetscErrorCode FNView(SlepcFN,PetscViewer)
    PetscErrorCode FNDestroy(SlepcFN*)
    PetscErrorCode FNReset(SlepcFN)
    PetscErrorCode FNSetType(SlepcFN,SlepcFNType)
    PetscErrorCode FNGetType(SlepcFN,SlepcFNType*)

    PetscErrorCode FNSetOptionsPrefix(SlepcFN,char[])
    PetscErrorCode FNGetOptionsPrefix(SlepcFN,char*[])
    PetscErrorCode FNAppendOptionsPrefix(SlepcFN,char[])
    PetscErrorCode FNSetFromOptions(SlepcFN)
    PetscErrorCode FNDuplicate(SlepcFN,MPI_Comm,SlepcFN*)

    PetscErrorCode FNSetScale(SlepcFN,PetscScalar,PetscScalar)
    PetscErrorCode FNGetScale(SlepcFN,PetscScalar*,PetscScalar*)
    PetscErrorCode FNSetMethod(SlepcFN,PetscInt)
    PetscErrorCode FNGetMethod(SlepcFN,PetscInt*)
    PetscErrorCode FNSetParallel(SlepcFN,SlepcFNParallelType)
    PetscErrorCode FNGetParallel(SlepcFN,SlepcFNParallelType*)
    PetscErrorCode FNEvaluateFunction(SlepcFN,PetscScalar,PetscScalar*)
    PetscErrorCode FNEvaluateDerivative(SlepcFN,PetscScalar,PetscScalar*)
    PetscErrorCode FNEvaluateFunctionMat(SlepcFN,PetscMat,PetscMat)
    PetscErrorCode FNEvaluateFunctionMatVec(SlepcFN,PetscMat,PetscVec)

    PetscErrorCode FNRationalSetNumerator(SlepcFN,PetscInt,PetscScalar[])
    PetscErrorCode FNRationalGetNumerator(SlepcFN,PetscInt*,PetscScalar*[])
    PetscErrorCode FNRationalSetDenominator(SlepcFN,PetscInt,PetscScalar[])
    PetscErrorCode FNRationalGetDenominator(SlepcFN,PetscInt*,PetscScalar*[])

    PetscErrorCode FNCombineSetChildren(SlepcFN,SlepcFNCombineType,SlepcFN,SlepcFN)
    PetscErrorCode FNCombineGetChildren(SlepcFN,SlepcFNCombineType*,SlepcFN*,SlepcFN*)

    PetscErrorCode FNPhiSetIndex(SlepcFN,PetscInt)
    PetscErrorCode FNPhiGetIndex(SlepcFN,PetscInt*)

# --------------------------------------------------------------------

# unary operations

cdef FN fn_pos(FN self):
    cdef FN fn = type(self)()
    cdef MPI_Comm comm = def_Comm(None, PetscObjectComm(<PetscObject>self.fn))
    CHKERR(FNDuplicate(self.fn, comm, &fn.fn))
    return fn

cdef FN fn_neg(FN self):
    cdef PetscScalar alpha = 1, beta = 1
    cdef FN fn = <FN> fn_pos(self)
    CHKERR(FNGetScale(fn.fn, &alpha, &beta))
    CHKERR(FNSetScale(fn.fn, alpha, -beta))
    return fn

# inplace binary operations

cdef FN fn_iadd(FN self, other):
    cdef PetscScalar alpha = 1
    cdef FN comb = type(self)()
    cdef SlepcFN fn = NULL
    cdef MPI_Comm comm = def_Comm(None, PetscObjectComm(<PetscObject>self.fn))
    CHKERR(FNCreate(comm, &comb.fn))
    CHKERR(FNSetType(comb.fn, FNCOMBINE))
    CHKERR(FNSetFromOptions(comb.fn))
    if isinstance(other, FN):
        fn = (<FN>other).fn
        CHKERR(PetscObjectReference(<PetscObject>fn))
    else:
        alpha = asScalar(other)
        CHKERR(FNCreate(comm, &fn))
        CHKERR(FNSetType(fn, FNRATIONAL))
        CHKERR(FNSetFromOptions(fn))
        CHKERR(FNRationalSetNumerator(fn, 1, &alpha));
    CHKERR(FNCombineSetChildren(comb.fn, FN_COMBINE_ADD, self.fn, fn))
    CHKERR(FNDestroy(&fn))
    return comb

cdef FN fn_isub(FN self, other):
    cdef PetscScalar alpha = 1, beta = 1
    cdef FN comb = type(self)()
    cdef SlepcFN fn = NULL
    cdef MPI_Comm comm = def_Comm(None, PetscObjectComm(<PetscObject>self.fn))
    CHKERR(FNCreate(comm, &comb.fn))
    CHKERR(FNSetType(comb.fn, FNCOMBINE))
    CHKERR(FNSetFromOptions(comb.fn))
    if isinstance(other, FN):
        CHKERR(FNDuplicate((<FN>other).fn, comm, &fn))
        CHKERR(FNGetScale(fn, &alpha, &beta))
        CHKERR(FNSetScale(fn, alpha, -beta))
    else:
        alpha = -asScalar(other)
        CHKERR(FNCreate(comm, &fn))
        CHKERR(FNSetType(fn, FNRATIONAL))
        CHKERR(FNSetFromOptions(fn))
        CHKERR(FNRationalSetNumerator(fn, 1, &alpha));
    CHKERR(FNCombineSetChildren(comb.fn, FN_COMBINE_ADD, self.fn, fn))
    CHKERR(FNDestroy(&fn))
    return comb

cdef FN fn_imul(FN self, other):
    cdef PetscScalar alpha = 1, beta = 1, gamma = 1
    cdef FN comb
    cdef MPI_Comm comm = def_Comm(None, PetscObjectComm(<PetscObject>self.fn))
    if isinstance(other, FN):
        comb = type(self)()
        CHKERR(FNCreate(comm, &comb.fn))
        CHKERR(FNSetType(comb.fn, FNCOMBINE))
        CHKERR(FNSetFromOptions(comb.fn))
        CHKERR(FNCombineSetChildren(comb.fn, FN_COMBINE_MULTIPLY, self.fn, (<FN>other).fn))
        return comb
    else:
        gamma = asScalar(other)
        CHKERR(FNGetScale(self.fn, &alpha, &beta))
        CHKERR(FNSetScale(self.fn, alpha, beta*gamma))
        return self

cdef FN fn_idiv(FN self, other):
    cdef PetscScalar alpha = 1, beta = 1, gamma = 1
    cdef FN comb
    cdef MPI_Comm comm = def_Comm(None, PetscObjectComm(<PetscObject>self.fn))
    if isinstance(other, FN):
        comb = type(self)()
        CHKERR(FNCreate(comm, &comb.fn))
        CHKERR(FNSetType(comb.fn, FNCOMBINE))
        CHKERR(FNSetFromOptions(comb.fn))
        CHKERR(FNCombineSetChildren(comb.fn, FN_COMBINE_DIVIDE, self.fn, (<FN>other).fn))
        return comb
    else:
        gamma = asScalar(other)
        CHKERR(FNGetScale(self.fn, &alpha, &beta))
        CHKERR(FNSetScale(self.fn, alpha, beta/gamma))
        return self

# binary operations

cdef FN fn_add(FN self, other):
    return fn_iadd(fn_pos(self), other)

cdef FN fn_sub(FN self, other):
    return fn_isub(fn_pos(self), other)

cdef FN fn_mul(FN self, other):
    return fn_imul(fn_pos(self), other)

cdef FN fn_div(FN self, other):
    return fn_idiv(fn_pos(self), other)

# reflected binary operations

cdef FN fn_radd(FN self, other):
    return fn_add(self, other)

cdef FN fn_rsub(FN self, other):
    cdef FN fn = <FN> fn_sub(self, other)
    CHKERR(FNSetScale(fn.fn, 1, -1))
    return fn

cdef FN fn_rmul(FN self, other):
    return fn_mul(self, other)

cdef FN fn_rdiv(FN self, other):
    <void>self; <void>other # unused
    return NotImplemented

# composition

cdef FN fn_matmul(FN self, other):
    cdef FN fn, comb
    cdef MPI_Comm comm = def_Comm(None, PetscObjectComm(<PetscObject>self.fn))
    if isinstance(other, FN):
        comb = type(self)()
        CHKERR(FNCreate(comm, &comb.fn))
        CHKERR(FNSetType(comb.fn, FNCOMBINE))
        CHKERR(FNSetFromOptions(comb.fn))
        CHKERR(FNCombineSetChildren(comb.fn, FN_COMBINE_COMPOSE, (<FN>other).fn, self.fn))
        return comb
    else:
        return NotImplemented




# -----------------------------------------------------------------------------

class LMEType(object):
    """
    LME type.

    - `KRYLOV`:  Restarted Krylov solver.
    """
    KRYLOV   = S_(LMEKRYLOV)

class LMEConvergedReason(object):
    """
    LME convergence reasons.

    - `CONVERGED_TOL`:       All eigenpairs converged to requested tolerance.
    - `DIVERGED_ITS`:        Maximum number of iterations exceeded.
    - `DIVERGED_BREAKDOWN`:  Solver failed due to breakdown.
    - `CONVERGED_ITERATING`: Iteration not finished yet.
    """
    CONVERGED_TOL       = LME_CONVERGED_TOL
    DIVERGED_ITS        = LME_DIVERGED_ITS
    DIVERGED_BREAKDOWN  = LME_DIVERGED_BREAKDOWN
    CONVERGED_ITERATING = LME_CONVERGED_ITERATING
    ITERATING           = LME_CONVERGED_ITERATING

class LMEProblemType(object):
    """
    LME problem type.

    - `LYAPUNOV`:      Continuous-time Lyapunov.
    - `SYLVESTER`:     Continuous-time Sylvester.
    - `GEN_LYAPUNOV`:  Generalized Lyapunov.
    - `GEN_SYLVESTER`: Generalized Sylvester.
    - `DT_LYAPUNOV`:   Discrete-time Lyapunov.
    - `STEIN`:         Stein.
    """
    LYAPUNOV        =  LME_LYAPUNOV
    SYLVESTER       =  LME_SYLVESTER
    GEN_LYAPUNOV    =  LME_GEN_LYAPUNOV
    GEN_SYLVESTER   =  LME_GEN_SYLVESTER
    DT_LYAPUNOV     =  LME_DT_LYAPUNOV
    STEIN           =  LME_STEIN

# -----------------------------------------------------------------------------

cdef class LME(Object):

    """LME."""

    Type            = LMEType
    ProblemType     = LMEProblemType
    ConvergedReason = LMEConvergedReason

    def __cinit__(self):
        self.obj = <PetscObject*> &self.lme
        self.lme = NULL

    def view(self, Viewer viewer=None) -> None:
        """
        Print the LME data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.
        """
        cdef PetscViewer vwr = def_Viewer(viewer)
        CHKERR( LMEView(self.lme, vwr) )

    def destroy(self) -> Self:
        """
        Destroy the LME object.

        Collective.
        """
        CHKERR( LMEDestroy(&self.lme) )
        self.lme = NULL
        return self

    def reset(self) -> None:
        """
        Reset the LME object.

        Collective.
        """
        CHKERR( LMEReset(self.lme) )

    def create(self, comm: Comm | None = None) -> Self:
        """
        Create the LME object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator. If not provided, it defaults to all processes.
        """
        cdef MPI_Comm ccomm = def_Comm(comm, SLEPC_COMM_DEFAULT())
        cdef SlepcLME newlme = NULL
        CHKERR( LMECreate(ccomm, &newlme) )
        CHKERR( SlepcCLEAR(self.obj) ); self.lme = newlme
        return self

    def setType(self, lme_type: Type | str) -> None:
        """
        Set the particular solver to be used in the LME object.

        Logically collective.

        Parameters
        ----------
        lme_type
            The solver to be used.
        """
        cdef SlepcLMEType cval = NULL
        lme_type = str2bytes(lme_type, &cval)
        CHKERR( LMESetType(self.lme, cval) )

    def getType(self) -> str:
        """
        Get the LME type of this object.

        Not collective.

        Returns
        -------
        str
            The solver currently being used.
        """
        cdef SlepcLMEType lme_type = NULL
        CHKERR( LMEGetType(self.lme, &lme_type) )
        return bytes2str(lme_type)

    def setProblemType(self, lme_problem_type: ProblemType | str) -> None:
        """
        Set the LME problem type of this object.

        Logically collective.

        Parameters
        ----------
        lme_problem_type
            The problem type to be used.
        """
        cdef SlepcLMEProblemType val = lme_problem_type
        CHKERR( LMESetProblemType(self.lme, val) )

    def getProblemType(self) -> ProblemType:
        """
        Get the LME problem type of this object.

        Not collective.

        Returns
        -------
        ProblemType
            The problem type currently being used.
        """
        cdef SlepcLMEProblemType val = LME_LYAPUNOV
        CHKERR( LMEGetProblemType(self.lme, &val))
        return val

    def setCoefficients(self, Mat A, Mat B = None, Mat D = None, Mat E = None) -> None:
        """
        Set the coefficient matrices.

        Collective.

        Set the coefficient matrices that define the linear matrix equation
        to be solved.

        Parameters
        ----------
        A
            First coefficient matrix
        B
            Second coefficient matrix, optional
        D
            Third coefficient matrix, optional
        E
            Fourth coefficient matrix, optional
        """
        cdef PetscMat Amat = A.mat
        cdef PetscMat Bmat = B.mat if B is not None else <PetscMat>NULL
        cdef PetscMat Dmat = D.mat if D is not None else <PetscMat>NULL
        cdef PetscMat Emat = E.mat if E is not None else <PetscMat>NULL
        CHKERR( LMESetCoefficients(self.lme, Amat, Bmat, Dmat, Emat))

    def getCoefficients(self) -> tuple[Mat, Mat | None, Mat | None, Mat | None]:
        """
        Get the coefficient matrices of the matrix equation.

        Collective.

        Returns
        -------
        ``A``
            First coefficient matrix
        ``B``
            Second coefficient matrix, if available
        ``D``
            Third coefficient matrix, if available
        ``E``
            Fourth coefficient matrix, if available
        """
        cdef PetscMat Amat, Bmat, Dmat, Emat
        cdef Mat A = Mat(), B = None, D = None, E = None
        CHKERR( LMEGetCoefficients(self.lme, &Amat, &Bmat, &Dmat, &Emat) )
        A.mat = Amat
        CHKERR( PetscINCREF(A.obj) )
        if Bmat:
            B = Mat(); B.mat = Bmat; CHKERR( PetscINCREF(B.obj) )
        if Dmat:
            D = Mat(); D.mat = Dmat; CHKERR( PetscINCREF(D.obj) )
        if Emat:
            E = Mat(); E.mat = Emat; CHKERR( PetscINCREF(E.obj) )
        return (A, B, D, E)

    def setRHS(self, Mat C) -> None:
        """
        Set the right-hand side of the matrix equation.

        Collective.

        Set the right-hand side of the matrix equation, as a low-rank
        matrix.

        Parameters
        ----------
        C
            The right-hand side matrix
        """
        CHKERR( LMESetRHS(self.lme, C.mat) )

    def getRHS(self) -> Mat:
        """
        Get the right-hand side of the matrix equation.

        Collective.

        Returns
        -------
        C
            The low-rank matrix
        """
        cdef Mat C = Mat()
        CHKERR( LMEGetRHS(self.lme, &C.mat) )
        CHKERR( PetscINCREF(C.obj) )
        return C

    def setSolution(self, Mat X) -> None:
        """
        Set the placeholder for the solution of the matrix equation.

        Collective.

        Set the placeholder for the solution of the matrix
        equation, as a low-rank matrix.

        Parameters
        ----------
        X
            The solution matrix
        """
        CHKERR( LMESetSolution(self.lme, X.mat) )

    def getSolution(self) -> Mat:
        """
        Get the solution of the matrix equation.

        Collective.

        Returns
        -------
        X
            The low-rank matrix
        """
        cdef Mat X = Mat()
        CHKERR( LMEGetSolution(self.lme, &X.mat) )
        CHKERR( PetscINCREF(X.obj) )
        return X

    def getErrorEstimate(self) -> float:
        """
        Get the error estimate obtained during solve.

        Not collective.

        Returns
        -------
        float
            The error estimate
        """
        cdef PetscReal rval = 0
        CHKERR( LMEGetErrorEstimate(self.lme, &rval) )
        return toReal(rval)

    def computeError(self) -> float:
        """
        Compute the error associated with the last equation solved.

        Collective.

        Computes the error (based on the residual norm) associated with the
        last equation solved.

        Returns
        -------
        float
            The error
        """
        cdef PetscReal rval = 0
        CHKERR( LMEComputeError(self.lme, &rval) )
        return toReal(rval)

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for searching for all LME options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this LME object.
        """
        cdef const char *prefix = NULL
        CHKERR( LMEGetOptionsPrefix(self.lme, &prefix) )
        return bytes2str(prefix)

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for searching for all LME options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all LME option requests.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( LMESetOptionsPrefix(self.lme, cval) )

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for searching in the database.

        Logically collective.

        Append to the prefix used for searching for all LME options in the
        database.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all LME option requests.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( LMEAppendOptionsPrefix(self.lme, cval) )

    def setFromOptions(self) -> None:
        """
        Set LME options from the options database.

        Collective.

        Sets LME options from the options database. This routine must
        be called before `setUp()` if the user is to be allowed to set
        the solver type.
        """
        CHKERR( LMESetFromOptions(self.lme) )

    def getTolerances(self) -> tuple[float, int]:
        """
        Get the tolerance and maximum iteration count.

        Not collective.

        Get the tolerance and maximum iteration count used by the
        default LME convergence tests.

        Returns
        -------
        tol: float
            The convergence tolerance.
        max_it: int
            The maximum number of iterations
        """
        cdef PetscReal rval = 0
        cdef PetscInt  ival = 0
        CHKERR( LMEGetTolerances(self.lme, &rval, &ival) )
        return (toReal(rval), toInt(ival))

    def setTolerances(self, tol: float | None = None, max_it: int | None = None) -> None:
        """
        Set the tolerance and maximum iteration count.

        Logically collective.

        Set the tolerance and maximum iteration count used by the
        default LME convergence tests.

        Parameters
        ----------
        tol
            The convergence tolerance.
        max_it
            The maximum number of iterations
        """
        cdef PetscReal rval = PETSC_DEFAULT
        cdef PetscInt  ival = PETSC_DEFAULT
        if tol    is not None: rval = asReal(tol)
        if max_it is not None: ival = asInt(max_it)
        CHKERR( LMESetTolerances(self.lme, rval, ival) )

    def getDimensions(self) -> int:
        """
        Get the dimension of the subspace used by the solver.

        Not collective.

        Returns
        -------
        int
            Maximum dimension of the subspace to be used by the solver.
        """
        cdef PetscInt ival = 0
        CHKERR( LMEGetDimensions(self.lme, &ival) )
        return toInt(ival)

    def setDimensions(self, ncv: int) -> None:
        """
        Set the dimension of the subspace to be used by the solver.

        Logically collective.

        Parameters
        ----------
        ncv
            Maximum dimension of the subspace to be used by the solver.
        """
        cdef PetscInt ival = asInt(ncv)
        CHKERR( LMESetDimensions(self.lme, ival) )

    def getBV(self) -> BV:
        """
        Get the basis vector object associated to the LME object.

        Not collective.

        Returns
        -------
        BV
            The basis vectors context.
        """
        cdef BV bv = BV()
        CHKERR( LMEGetBV(self.lme, &bv.bv) )
        CHKERR( PetscINCREF(bv.obj) )
        return bv

    def setBV(self, BV bv) -> None:
        """
        Set a basis vector object to the LME object.

        Collective.

        Parameters
        ----------
        bv
            The basis vectors context.
        """
        CHKERR( LMESetBV(self.lme, bv.bv) )

    def setMonitor(
        self,
        monitor: LMEMonitorFunction | None,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Append a monitor function to the list of monitors.

        Logically collective.
        """
        if monitor is None: return
        cdef object monitorlist = self.get_attr('__monitor__')
        if monitorlist is None:
            monitorlist = []
            self.set_attr('__monitor__', monitorlist)
            CHKERR( LMEMonitorSet(self.lme, LME_Monitor, NULL, NULL) )
        if args is None: args = ()
        if kargs is None: kargs = {}
        monitorlist.append((monitor, args, kargs))

    def getMonitor(self) -> LMEMonitorFunction:
        """
        Get the list of monitor functions.
        """
        return self.get_attr('__monitor__')

    def cancelMonitor(self) -> None:
        """
        Clear all monitors for an `LME` object.
        """
        CHKERR( LMEMonitorCancel(self.lme) )
        self.set_attr('__monitor__', None)

    def setUp(self) -> None:
        """
        Set up all the internal necessary data structures.

        Collective.

        Set up all the internal data structures necessary for the
        execution of the eigensolver.
        """
        CHKERR( LMESetUp(self.lme) )

    def solve(self) -> None:
        """
        Solve the linear matrix equation.

        Collective.
        """
        CHKERR( LMESolve(self.lme) )

    def getIterationNumber(self) -> int:
        """
        Get the current iteration number.

        Not collective.

        If the call to `solve()` is complete, then it returns the number of
        iterations carried out by the solution method.

        Returns
        -------
        int
            Iteration number.
        """
        cdef PetscInt ival = 0
        CHKERR( LMEGetIterationNumber(self.lme, &ival) )
        return toInt(ival)

    def getConvergedReason(self) -> ConvergedReason:
        """
        Get the reason why the `solve()` iteration was stopped.

        Not collective.

        Returns
        -------
        ConvergedReason
            Negative value indicates diverged, positive value converged.
        """
        cdef SlepcLMEConvergedReason val = LME_CONVERGED_ITERATING
        CHKERR( LMEGetConvergedReason(self.lme, &val) )
        return val

    def setErrorIfNotConverged(self, flg: bool = True) -> None:
        """
        Set `solve()` to generate an error if the solver has not converged.

        Logically collective.

        Parameters
        ----------
        flg
            True indicates you want the error generated.
        """
        cdef PetscBool tval = flg
        CHKERR( LMESetErrorIfNotConverged(self.lme, tval) )

    def getErrorIfNotConverged(self) -> bool:
        """
        Get if `solve()` generates an error if the solver does not converge.

        Not collective.

        Get a flag indicating whether `solve()` will generate an error if the
        solver does not converge.

        Returns
        -------
        bool
            True indicates you want the error generated.
        """
        cdef PetscBool tval = PETSC_FALSE
        CHKERR( LMEGetErrorIfNotConverged(self.lme, &tval) )
        return toBool(tval)

    #

    property tol:
        """The tolerance value used by the LME convergence tests."""
        def __get__(self) -> float:
            return self.getTolerances()[0]
        def __set__(self, value):
            self.setTolerances(tol=value)

    property max_it:
        """The maximum iteration count used by the LME convergence tests."""
        def __get__(self) -> int:
            return self.getTolerances()[1]
        def __set__(self, value):
            self.setTolerances(max_it=value)

    property fn:
        """The math function (FN) object associated to the LME object."""
        def __get__(self) -> FN:
            return self.getFN()
        def __set__(self, value):
            self.setFN(value)

    property bv:
        """The basis vectors (BV) object associated to the LME object."""
        def __get__(self) -> BV:
            return self.getBV()
        def __set__(self, value):
            self.setBV(value)

# -----------------------------------------------------------------------------

del LMEType
del LMEConvergedReason
del LMEProblemType

# -----------------------------------------------------------------------------



"""Type stubs for SLEPc ST module - spectral transformation."""

from enum import IntEnum, StrEnum

import petsc4py.PETSc
from petsc4py.PETSc import KSP, Comm, Mat, Vec, Viewer
from petsc4py.typing import Scalar

class STType(StrEnum):
    """ST types.

    - `SHELL`:   User-defined.
    - `SHIFT`:   Shift from origin.
    - `SINVERT`: Shift-and-invert.
    - `CAYLEY`:  Cayley transform.
    - `PRECOND`: Preconditioner.
    - `FILTER`:  Polynomial filter.
    """

    SHELL = ...
    """User-defined."""
    SHIFT = ...
    """Shift from origin."""
    SINVERT = ...
    """Shift-and-invert."""
    CAYLEY = ...
    """Cayley transform."""
    PRECOND = ...
    """Preconditioner."""
    FILTER = ...
    """Polynomial filter."""

class STMatMode(IntEnum):
    """ST matrix mode.

    - `COPY`:    A working copy of the matrix is created.
    - `INPLACE`: The operation is computed in-place.
    - `SHELL`:   The matrix A - sigma B is handled as an
                 implicit matrix.
    """

    COPY = ...
    """A working copy of the matrix is created."""
    INPLACE = ...
    """The operation is computed in-place."""
    SHELL = ...
    """The matrix A - sigma B is handled as an implicit matrix."""

class STFilterType(IntEnum):
    """ST filter type.

    - `FILTLAN`:  An adapted implementation of the Filtered Lanczos Package.
    - `CHEBYSHEV`: A polynomial filter based on a truncated Chebyshev series.
    """

    FILTLAN = ...
    """An adapted implementation of the Filtered Lanczos Package."""
    CHEBYSHEV = ...
    """A polynomial filter based on a truncated Chebyshev series."""

class STFilterDamping(IntEnum):
    """ST filter damping.

    - `NONE`:    No damping
    - `JACKSON`: Jackson damping
    - `LANCZOS`: Lanczos damping
    - `FEJER`:   Fejer damping
    """

    NONE = ...
    """No damping."""
    JACKSON = ...
    """Jackson damping."""
    LANCZOS = ...
    """Lanczos damping."""
    FEJER = ...
    """Fejer damping."""

class ST:
    """ST: Spectral Transformation."""

    Type = STType
    MatMode = STMatMode
    FilterType = STFilterType
    FilterDamping = STFilterDamping

    # Properties
    shift: float
    transform: bool
    mat_mode: STMatMode
    mat_structure: Mat.Structure
    ksp: KSP

    def __init__(self) -> None: ...
    def view(self, viewer: Viewer | None = None) -> None:
        """
        Print the ST data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.
        """
        ...

    def destroy(self) -> ST:
        """
        Destroy the ST object.

        Collective.
        """
        ...

    def reset(self) -> None:
        """
        Reset the ST object.

        Collective.
        """
        ...

    def create(self, comm: Comm | None = None) -> ST:
        """
        Create the ST object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to all processes.
        """
        ...

    def setType(self, st_type: STType | str) -> None:
        """
        Set the particular spectral transformation to be used.

        Logically collective.

        Parameters
        ----------
        st_type
            The spectral transformation to be used.
        """
        ...

    def getType(self) -> str:
        """
        Get the ST type of this object.

        Not collective.

        Returns
        -------
        str
            The spectral transformation currently being used.
        """
        ...

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for searching for all ST options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all ST option requests.
        """
        ...

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for searching for all ST options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this ST object.
        """
        ...

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for searching for all ST options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all ST option requests.
        """
        ...

    def setFromOptions(self) -> None:
        """
        Set ST options from the options database.

        Collective.
        """
        ...

    def setShift(self, shift: Scalar) -> None:
        """
        Set the shift associated with the spectral transformation.

        Collective.

        Parameters
        ----------
        shift
            The value of the shift.
        """
        ...

    def getShift(self) -> Scalar:
        """
        Get the shift associated with the spectral transformation.

        Not collective.

        Returns
        -------
        Scalar
            The value of the shift.
        """
        ...

    def setTransform(self, flag: bool = True) -> None:
        """
        Set a flag to indicate whether the transformed matrices are computed or not.

        Logically collective.

        Parameters
        ----------
        flag
            This flag is intended for the case of polynomial
            eigenproblems solved via linearization.
        """
        ...

    def getTransform(self) -> bool:
        """
        Get the flag indicating whether the transformed matrices are computed or not.

        Not collective.

        Returns
        -------
        bool
            The transform flag.
        """
        ...

    def setMatMode(self, mode: STMatMode) -> None:
        """
        Set a flag to indicate how the matrix is being shifted.

        Logically collective.

        Parameters
        ----------
        mode
            The mode flag.
        """
        ...

    def getMatMode(self) -> STMatMode:
        """
        Get a flag that indicates how the matrix is being shifted.

        Not collective.

        Returns
        -------
        STMatMode
            The mode flag.
        """
        ...

    def setMatrices(self, operators: list[Mat]) -> None:
        """
        Set the matrices associated with the eigenvalue problem.

        Collective.

        Parameters
        ----------
        operators
            The matrices associated with the eigensystem.
        """
        ...

    def getMatrices(self) -> list[Mat]:
        """
        Get the matrices associated with the eigenvalue problem.

        Collective.

        Returns
        -------
        list of petsc4py.PETSc.Mat
            The matrices associated with the eigensystem.
        """
        ...

    def setMatStructure(self, structure: Mat.Structure) -> None:
        """
        Set an internal Mat.Structure attribute.

        Logically collective.

        Parameters
        ----------
        structure
            Either same, different, or a subset of the non-zero
            sparsity pattern.
        """
        ...

    def getMatStructure(self) -> Mat.Structure:
        """
        Get the internal Mat.Structure attribute.

        Not collective.

        Returns
        -------
        petsc4py.PETSc.Mat.Structure
            The structure flag.
        """
        ...

    def setKSP(self, ksp: KSP) -> None:
        """
        Set the ``KSP`` object associated with the spectral transformation.

        Collective.

        Parameters
        ----------
        ksp
            The linear solver object.
        """
        ...

    def getKSP(self) -> KSP:
        """
        Get the ``KSP`` object associated with the spectral transformation.

        Collective.

        Returns
        -------
        `petsc4py.PETSc.KSP`
            The linear solver object.
        """
        ...

    def setPreconditionerMat(self, P: Mat | None = None) -> None:
        """
        Set the matrix to be used to build the preconditioner.

        Collective.

        Parameters
        ----------
        P
            The matrix that will be used in constructing the preconditioner.
        """
        ...

    def getPreconditionerMat(self) -> Mat:
        """
        Get the matrix previously set by setPreconditionerMat().

        Not collective.

        Returns
        -------
        petsc4py.PETSc.Mat
            The matrix that will be used in constructing the preconditioner.
        """
        ...

    def setSplitPreconditioner(
        self,
        operators: list[Mat],
        structure: petsc4py.PETSc.Mat.Structure | None = None,
    ) -> None:
        """
        Set the matrices to be used to build the preconditioner.

        Collective.

        Parameters
        ----------
        operators
            The matrices associated with the preconditioner.
        structure
            Either same, different, or a subset of the non-zero sparsity pattern.
        """
        ...

    def getSplitPreconditioner(
        self,
    ) -> tuple[list[Mat], petsc4py.PETSc.Mat.Structure]:
        """
        Get the matrices to be used to build the preconditioner.

        Not collective.

        Returns
        -------
        list of petsc4py.PETSc.Mat
            The list of matrices associated with the preconditioner.
        petsc4py.PETSc.Mat.Structure
            The structure flag.
        """
        ...

    def setUp(self) -> None:
        """
        Prepare for the use of a spectral transformation.

        Collective.
        """
        ...

    def apply(self, x: Vec, y: Vec) -> None:
        """
        Apply the spectral transformation operator to a vector.

        Collective.

        Parameters
        ----------
        x
            The input vector.
        y
            The result vector.
        """
        ...

    def applyTranspose(self, x: Vec, y: Vec) -> None:
        """
        Apply the transpose of the operator to a vector.

        Collective.

        Parameters
        ----------
        x
            The input vector.
        y
            The result vector.
        """
        ...

    def applyHermitianTranspose(self, x: Vec, y: Vec) -> None:
        """
        Apply the hermitian-transpose of the operator to a vector.

        Collective.

        Parameters
        ----------
        x
            The input vector.
        y
            The result vector.
        """
        ...

    def applyMat(self, x: Mat, y: Mat) -> None:
        """
        Apply the spectral transformation operator to a matrix.

        Collective.

        Parameters
        ----------
        x
            The input matrix.
        y
            The result matrix.
        """
        ...

    def getOperator(self) -> Mat:
        """
        Get a shell matrix that represents the operator of the spectral transformation.

        Collective.

        Returns
        -------
        petsc4py.PETSc.Mat
            Operator matrix.
        """
        ...

    def restoreOperator(self, op: Mat) -> None:
        """
        Restore the previously seized operator matrix.

        Logically collective.

        Parameters
        ----------
        op
            Operator matrix previously obtained with getOperator().
        """
        ...

    # Cayley-specific methods
    def setCayleyAntishift(self, tau: Scalar) -> None:
        """
        Set the value of the anti-shift for the Cayley spectral transformation.

        Logically collective.

        Parameters
        ----------
        tau
            The anti-shift.
        """
        ...

    def getCayleyAntishift(self) -> Scalar:
        """
        Get the value of the anti-shift for the Cayley spectral transformation.

        Not collective.

        Returns
        -------
        Scalar
            The anti-shift.
        """
        ...

    # Filter-specific methods
    def setFilterType(self, filter_type: STFilterType) -> None:
        """
        Set the method to be used to build the polynomial filter.

        Logically collective.

        Parameters
        ----------
        filter_type
            The type of filter.
        """
        ...

    def getFilterType(self) -> STFilterType:
        """
        Get the method to be used to build the polynomial filter.

        Not collective.

        Returns
        -------
        STFilterType
            The type of filter.
        """
        ...

    def setFilterInterval(self, inta: float, intb: float) -> None:
        """
        Set the interval containing the desired eigenvalues.

        Logically collective.

        Parameters
        ----------
        inta
            The left end of the interval.
        intb
            The right end of the interval.
        """
        ...

    def getFilterInterval(self) -> tuple[float, float]:
        """
        Get the interval containing the desired eigenvalues.

        Not collective.

        Returns
        -------
        inta: float
            The left end of the interval.
        intb: float
            The right end of the interval.
        """
        ...

    def setFilterRange(self, left: float, right: float) -> None:
        """
        Set the numerical range (or field of values) of the matrix.

        Logically collective.

        Parameters
        ----------
        left
            The left end of the interval.
        right
            The right end of the interval.
        """
        ...

    def getFilterRange(self) -> tuple[float, float]:
        """
        Get the interval containing all eigenvalues.

        Not collective.

        Returns
        -------
        left: float
            The left end of the interval.
        right: float
            The right end of the interval.
        """
        ...

    def setFilterDegree(self, deg: int) -> None:
        """
        Set the degree of the filter polynomial.

        Logically collective.

        Parameters
        ----------
        deg
            The polynomial degree.
        """
        ...

    def getFilterDegree(self) -> int:
        """
        Get the degree of the filter polynomial.

        Not collective.

        Returns
        -------
        int
            The polynomial degree.
        """
        ...

    def setFilterDamping(self, damping: STFilterDamping) -> None:
        """
        Set the type of damping to be used in the polynomial filter.

        Logically collective.

        Parameters
        ----------
        damping
            The type of damping.
        """
        ...

    def getFilterDamping(self) -> STFilterDamping:
        """
        Get the type of damping used in the polynomial filter.

        Not collective.

        Returns
        -------
        STFilterDamping
            The type of damping.
        """
        ...



"""Type stubs for SLEPc BV module."""

from enum import IntEnum, StrEnum
from typing import Any, Self, Sequence

from petsc4py.PETSc import Comm, Mat, Random, Vec, Viewer
from petsc4py.typing import ArrayScalar, LayoutSizeSpec, NormTypeSpec, Scalar

class BVType(StrEnum):
    """BV type."""

    MAT = ...
    """Matrix-based BV."""
    SVEC = ...
    """Single-vector BV."""
    VECS = ...
    """Vector array BV."""
    CONTIGUOUS = ...
    """Contiguous memory BV."""
    TENSOR = ...
    """Tensor product BV."""

class BVOrthogType(IntEnum):
    """BV orthogonalization types.

    - `CGS`: Classical Gram-Schmidt.
    - `MGS`: Modified Gram-Schmidt.
    """

    CGS = ...
    """Classical Gram-Schmidt."""
    MGS = ...
    """Modified Gram-Schmidt."""

class BVOrthogRefineType(IntEnum):
    """BV orthogonalization refinement types.

    - `IFNEEDED`: Reorthogonalize if a criterion is satisfied.
    - `NEVER`:    Never reorthogonalize.
    - `ALWAYS`:   Always reorthogonalize.
    """

    IFNEEDED = ...
    """Reorthogonalize if a criterion is satisfied."""
    NEVER = ...
    """Never reorthogonalize."""
    ALWAYS = ...
    """Always reorthogonalize."""

class BVOrthogBlockType(IntEnum):
    """BV block-orthogonalization types.

    - `GS`:       Gram-Schmidt.
    - `CHOL`:     Cholesky.
    - `TSQR`:     Tall-skinny QR.
    - `TSQRCHOL`: Tall-skinny QR with Cholesky.
    - `SVQB`:     SVQB.
    """

    GS = ...
    """Gram-Schmidt."""
    CHOL = ...
    """Cholesky."""
    TSQR = ...
    """Tall-skinny QR."""
    TSQRCHOL = ...
    """Tall-skinny QR with Cholesky."""
    SVQB = ...
    """SVQB."""

class BVMatMultType(IntEnum):
    """BV mat-mult types.

    - `VECS`: Perform a matrix-vector multiply per each column.
    - `MAT`:  Carry out a Mat-Mat product with a dense matrix.
    """

    VECS = ...
    """Perform a matrix-vector multiply per each column."""
    MAT = ...
    """Carry out a Mat-Mat product with a dense matrix."""

class BVSVDMethod(IntEnum):
    """BV methods for computing the SVD.

    - `REFINE`: Based on the SVD of the cross product matrix S^H S, with refinement.
    - `QR`:     Based on the SVD of the triangular factor of qr(S).
    - `QR_CAA`: Variant of QR intended for use in communication-avoiding Arnoldi.
    """

    REFINE = ...
    """Based on the SVD of the cross product matrix S^H S, with refinement."""
    QR = ...
    """Based on the SVD of the triangular factor of qr(S)."""
    QR_CAA = ...
    """Variant of QR intended for use in communication-avoiding Arnoldi."""

class BV:
    """Basis vectors object.

    BV is the basic SLEPc object used to represent a collection of
    vectors, typically used as a basis in spectral projectors.

    See Also
    --------
    slepc.BV
    """

    Type = BVType
    OrthogType = BVOrthogType
    OrthogRefineType = BVOrthogRefineType
    OrthogBlockType = BVOrthogBlockType
    MatMultType = BVMatMultType
    SVDMethod = BVSVDMethod

    # --- Unary operations ---

    def __pos__(self) -> BV: ...
    def __neg__(self) -> BV: ...

    # --- In-place binary operations ---

    def __iadd__(self, other: BV | Scalar) -> BV: ...
    def __isub__(self, other: BV | Scalar) -> BV: ...
    def __imul__(self, other: BV | Scalar) -> BV: ...
    def __idiv__(self, other: BV | Scalar) -> BV: ...
    def __itruediv__(self, other: BV | Scalar) -> BV: ...

    # --- Binary operations ---

    def __add__(self, other: BV | Scalar) -> BV: ...
    def __radd__(self, other: BV | Scalar) -> BV: ...
    def __sub__(self, other: BV | Scalar) -> BV: ...
    def __rsub__(self, other: BV | Scalar) -> BV: ...
    def __mul__(self, other: BV | Scalar) -> BV: ...
    def __rmul__(self, other: BV | Scalar) -> BV: ...
    def __div__(self, other: BV | Scalar) -> BV: ...
    def __rdiv__(self, other: BV | Scalar) -> BV: ...
    def __truediv__(self, other: BV | Scalar) -> BV: ...
    def __rtruediv__(self, other: BV | Scalar) -> BV: ...

    # --- View and Lifecycle ---

    def view(self, viewer: Viewer | None = None) -> None:
        """Print the BV data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.

        See Also
        --------
        slepc.BVView
        """
        ...

    def destroy(self) -> Self:
        """Destroy the BV object.

        Collective.

        See Also
        --------
        slepc.BVDestroy
        """
        ...

    def create(self, comm: Comm | None = None) -> Self:
        """Create the BV object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to all
            processes.

        See Also
        --------
        slepc.BVCreate
        """
        ...

    def createFromMat(self, A: Mat) -> Self:
        """Create a basis vectors object from a dense Mat object.

        Collective.

        Parameters
        ----------
        A
            A dense tall-skinny matrix.

        See Also
        --------
        slepc.BVCreateFromMat
        """
        ...

    def createMat(self) -> Mat:
        """Create a new Mat object of dense type and copy the contents of the BV.

        Collective.

        Returns
        -------
        Mat
            The new matrix.

        See Also
        --------
        slepc.BVCreateMat
        """
        ...

    def duplicate(self) -> BV:
        """Duplicate the BV object with the same type and dimensions.

        Collective.

        See Also
        --------
        slepc.BVDuplicate
        """
        ...

    def duplicateResize(self, m: int) -> BV:
        """Create a BV object of the same type and dimensions as an existing one.

        Collective.

        Parameters
        ----------
        m
            The number of columns.

        Notes
        -----
        With possibly different number of columns.

        See Also
        --------
        slepc.BVDuplicateResize
        """
        ...

    def copy(self, result: BV | None = None) -> BV:
        """Copy a basis vector object into another one.

        Logically collective.

        Parameters
        ----------
        result
            The copy.

        See Also
        --------
        slepc.BVCopy
        """
        ...

    # --- Type and Setup ---

    def setType(self, bv_type: BVType | str) -> None:
        """Set the type for the BV object.

        Logically collective.

        Parameters
        ----------
        bv_type
            The inner product type to be used.

        See Also
        --------
        slepc.BVSetType
        """
        ...

    def getType(self) -> str:
        """Get the BV type of this object.

        Not collective.

        Returns
        -------
        str
            The inner product type currently being used.

        See Also
        --------
        slepc.BVGetType
        """
        ...

    def setSizes(self, sizes: LayoutSizeSpec, m: int) -> None:
        """Set the local and global sizes, and the number of columns.

        Collective.

        Parameters
        ----------
        sizes
            The global size ``N`` or a two-tuple ``(n, N)``
            with the local and global sizes.
        m
            The number of columns.

        Notes
        -----
        Either ``n`` or ``N`` (but not both) can be ``PETSc.DECIDE``
        or ``None`` to have it automatically set.

        See Also
        --------
        slepc.BVSetSizes
        """
        ...

    def setSizesFromVec(self, w: Vec, m: int) -> None:
        """Set the local and global sizes, and the number of columns.

        Collective.

        Local and global sizes are specified indirectly by passing a template
        vector.

        Parameters
        ----------
        w
            The template vector.
        m
            The number of columns.

        See Also
        --------
        slepc.BVSetSizesFromVec
        """
        ...

    def getSizes(self) -> tuple[tuple[int, int], int]:
        """Get the local and global sizes, and the number of columns.

        Not collective.

        Returns
        -------
        sizes : tuple[int, int]
            The local and global sizes (n, N).
        m : int
            The number of columns.

        See Also
        --------
        slepc.BVGetSizes
        """
        ...

    def setLeadingDimension(self, ld: int) -> None:
        """Set the leading dimension.

        Not collective.

        Parameters
        ----------
        ld
            The leading dimension.

        See Also
        --------
        slepc.BVSetLeadingDimension
        """
        ...

    def getLeadingDimension(self) -> int:
        """Get the leading dimension.

        Not collective.

        Returns
        -------
        int
            The leading dimension.

        See Also
        --------
        slepc.BVGetLeadingDimension
        """
        ...

    # --- Options ---

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """Set the prefix used for searching for all BV options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all BV option requests.

        Notes
        -----
        A hyphen (``-``) must NOT be given at the beginning of the
        prefix name.  The first character of all runtime options is
        AUTOMATICALLY the hyphen.

        See Also
        --------
        slepc.BVSetOptionsPrefix
        """
        ...

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """Append to the prefix used for searching for all BV options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all BV option requests.

        See Also
        --------
        slepc.BVAppendOptionsPrefix
        """
        ...

    def getOptionsPrefix(self) -> str:
        """Get the prefix used for searching for all BV options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this BV object.

        See Also
        --------
        slepc.BVGetOptionsPrefix
        """
        ...

    def setFromOptions(self) -> None:
        """Set BV options from the options database.

        Collective.

        Notes
        -----
        To see all options, run your program with the ``-help``
        option.

        See Also
        --------
        slepc.BVSetFromOptions
        """
        ...

    # --- Orthogonalization ---

    def getOrthogonalization(
        self,
    ) -> tuple[BVOrthogType, BVOrthogRefineType, float, BVOrthogBlockType]:
        """Get the orthogonalization settings from the BV object.

        Not collective.

        Returns
        -------
        type : BVOrthogType
            The type of orthogonalization technique.
        refine : BVOrthogRefineType
            The type of refinement.
        eta : float
            Parameter for selective refinement (used when the
            refinement type is `BV.OrthogRefineType.IFNEEDED`).
        block : BVOrthogBlockType
            The type of block orthogonalization.

        See Also
        --------
        slepc.BVGetOrthogonalization
        """
        ...

    def setOrthogonalization(
        self,
        otype: BVOrthogType | None = None,
        refine: BVOrthogRefineType | None = None,
        eta: float | None = None,
        block: BVOrthogBlockType | None = None,
    ) -> None:
        """Set the method used for the (block-)orthogonalization of vectors.

        Logically collective.

        Orthogonalization of vectors (classical or modified Gram-Schmidt
        with or without refinement), and for the block-orthogonalization
        (simultaneous orthogonalization of a set of vectors).

        Parameters
        ----------
        otype
            The type of orthogonalization technique.
        refine
            The type of refinement.
        eta
            Parameter for selective refinement.
        block
            The type of block orthogonalization.

        Notes
        -----
        The default settings work well for most problems.

        The parameter ``eta`` should be a real value between ``0`` and
        ``1`` (or `DETERMINE`).  The value of ``eta`` is used only when
        the refinement type is `BV.OrthogRefineType.IFNEEDED`.

        When using several processors, `BV.OrthogType.MGS` is likely to
        result in bad scalability.

        If the method set for block orthogonalization is GS, then the
        computation is done column by column with the vector orthogonalization.

        See Also
        --------
        slepc.BVSetOrthogonalization
        """
        ...

    def getMatMultMethod(self) -> BVMatMultType:
        """Get the method used for the `matMult()` operation.

        Not collective.

        Returns
        -------
        BVMatMultType
            The method for the `matMult()` operation.

        See Also
        --------
        slepc.BVGetMatMultMethod
        """
        ...

    def setMatMultMethod(self, method: BVMatMultType) -> None:
        """Set the method used for the `matMult()` operation.

        Logically collective.

        Parameters
        ----------
        method
            The method for the `matMult()` operation.

        See Also
        --------
        slepc.BVSetMatMultMethod
        """
        ...

    # --- Matrix ---

    def getMatrix(self) -> tuple[Mat | None, bool]:
        """Get the matrix representation of the inner product.

        Not collective.

        Returns
        -------
        mat : Mat | None
            The matrix of the inner product.
        indef : bool
            Whether the matrix is indefinite.

        See Also
        --------
        slepc.BVGetMatrix
        """
        ...

    def setMatrix(self, mat: Mat | None, indef: bool = False) -> None:
        """Set the bilinear form to be used for inner products.

        Collective.

        Parameters
        ----------
        mat
            The matrix of the inner product.
        indef
            Whether the matrix is indefinite.

        See Also
        --------
        slepc.BVSetMatrix
        """
        ...

    def applyMatrix(self, x: Vec, y: Vec) -> None:
        """Multiply a vector with the matrix associated to the bilinear form.

        Neighbor-wise collective.

        Parameters
        ----------
        x
            The input vector.
        y
            The result vector.

        Notes
        -----
        If the bilinear form has no associated matrix this function
        copies the vector.

        See Also
        --------
        slepc.BVApplyMatrix
        """
        ...

    # --- Active Columns ---

    def setActiveColumns(self, l: int, k: int) -> None:
        """Set the columns that will be involved in operations.

        Logically collective.

        Parameters
        ----------
        l
            The leading number of columns.
        k
            The active number of columns.

        See Also
        --------
        slepc.BVSetActiveColumns
        """
        ...

    def getActiveColumns(self) -> tuple[int, int]:
        """Get the current active dimensions.

        Not collective.

        Returns
        -------
        l : int
            The leading number of columns.
        k : int
            The active number of columns.

        See Also
        --------
        slepc.BVGetActiveColumns
        """
        ...

    # --- Column Operations ---

    def scaleColumn(self, j: int, alpha: Scalar) -> None:
        """Scale column j by alpha.

        Logically collective.

        Parameters
        ----------
        j
            Column number to be scaled.
        alpha
            Scaling factor.

        See Also
        --------
        slepc.BVScaleColumn
        """
        ...

    def scale(self, alpha: Scalar) -> None:
        """Multiply the entries by a scalar value.

        Logically collective.

        Parameters
        ----------
        alpha
            Scaling factor.

        Notes
        -----
        All active columns (except the leading ones) are scaled.

        See Also
        --------
        slepc.BVScale
        """
        ...

    def insertVec(self, j: int, w: Vec) -> None:
        """Insert a vector into the specified column.

        Logically collective.

        Parameters
        ----------
        j
            The column to be overwritten.
        w
            The vector to be copied.

        See Also
        --------
        slepc.BVInsertVec
        """
        ...

    def insertVecs(self, s: int, W: Vec | list[Vec], orth: bool = False) -> int:
        """Insert a set of vectors into specified columns.

        Collective.

        Parameters
        ----------
        s
            The first column to be overwritten.
        W
            Set of vectors to be copied.
        orth
            Flag indicating if the vectors must be orthogonalized.

        Returns
        -------
        int
            Number of linearly independent vectors.

        Notes
        -----
        Copies the contents of vectors W into self(:,s:s+n), where n is the
        length of W. If orthogonalization flag is set then the vectors are
        copied one by one then orthogonalized against the previous one.  If any
        are linearly dependent then it is discarded and the value of m is
        decreased.

        See Also
        --------
        slepc.BVInsertVecs
        """
        ...

    def insertConstraints(self, C: Vec | list[Vec]) -> int:
        """Insert a set of vectors as constraints.

        Collective.

        Parameters
        ----------
        C
            Set of vectors to be inserted as constraints.

        Returns
        -------
        int
            Number of constraints.

        Notes
        -----
        The constraints are relevant only during orthogonalization. Constraint
        vectors span a subspace that is deflated in every orthogonalization
        operation, so they are intended for removing those directions from the
        orthogonal basis computed in regular BV columns.

        See Also
        --------
        slepc.BVInsertConstraints
        """
        ...

    def setNumConstraints(self, nc: int) -> None:
        """Set the number of constraints.

        Logically collective.

        Parameters
        ----------
        nc
            The number of constraints.

        See Also
        --------
        slepc.BVSetNumConstraints
        """
        ...

    def getNumConstraints(self) -> int:
        """Get the number of constraints.

        Not collective.

        Returns
        -------
        int
            The number of constraints.

        See Also
        --------
        slepc.BVGetNumConstraints
        """
        ...

    # --- Vector Operations ---

    def createVec(self) -> Vec:
        """Create a Vec with the type and dimensions of the columns of the BV.

        Collective.

        Returns
        -------
        Vec
            New vector.

        See Also
        --------
        slepc.BVCreateVec
        """
        ...

    def setVecType(self, vec_type: str) -> None:
        """Set the vector type.

        Collective.

        Parameters
        ----------
        vec_type
            Vector type used when creating vectors with `createVec`.

        See Also
        --------
        slepc.BVSetVecType
        """
        ...

    def getVecType(self) -> str:
        """Get the vector type used by the basis vectors object.

        Not collective.

        Returns
        -------
        str
            The vector type.

        See Also
        --------
        slepc.BVGetVecType
        """
        ...

    def copyVec(self, j: int, v: Vec) -> None:
        """Copy one of the columns of a basis vectors object into a Vec.

        Logically collective.

        Parameters
        ----------
        j
            The column number to be copied.
        v
            A vector.

        See Also
        --------
        slepc.BVCopyVec
        """
        ...

    def copyColumn(self, j: int, i: int) -> None:
        """Copy the values from one of the columns to another one.

        Logically collective.

        Parameters
        ----------
        j
            The number of the source column.
        i
            The number of the destination column.

        See Also
        --------
        slepc.BVCopyColumn
        """
        ...

    # --- Tolerance ---

    def setDefiniteTolerance(self, deftol: float) -> None:
        """Set the tolerance to be used when checking a definite inner product.

        Logically collective.

        Parameters
        ----------
        deftol
            The tolerance.

        See Also
        --------
        slepc.BVSetDefiniteTolerance
        """
        ...

    def getDefiniteTolerance(self) -> float:
        """Get the tolerance to be used when checking a definite inner product.

        Not collective.

        Returns
        -------
        float
            The tolerance.

        See Also
        --------
        slepc.BVGetDefiniteTolerance
        """
        ...

    # --- Dot Products ---

    def dotVec(self, v: Vec) -> ArrayScalar:
        """Dot products of a vector against all the column vectors of the BV.

        Collective.

        Parameters
        ----------
        v
            A vector.

        Returns
        -------
        ArrayScalar
            The computed values.

        Notes
        -----
        This is analogue to VecMDot(), but using BV to represent a collection
        of vectors. The result is m = X^H y, so m_i is equal to x_j^H y.
        Note that here X is transposed as opposed to BVDot().

        If a non-standard inner product has been specified with BVSetMatrix(),
        then the result is m = X^H B y.

        See Also
        --------
        slepc.BVDotVec
        """
        ...

    def dotColumn(self, j: int) -> ArrayScalar:
        """Dot products of a column against all the column vectors of a BV.

        Collective.

        Parameters
        ----------
        j
            The index of the column.

        Returns
        -------
        ArrayScalar
            The computed values.

        See Also
        --------
        slepc.BVDotColumn
        """
        ...

    def getColumn(self, j: int) -> Vec:
        """Get a Vec object with the entries of the column of the BV object.

        Logically collective.

        Parameters
        ----------
        j
            The index of the requested column.

        Returns
        -------
        Vec
            The vector containing the jth column.

        Notes
        -----
        Modifying the returned Vec will change the BV entries as well.

        See Also
        --------
        slepc.BVGetColumn
        """
        ...

    def restoreColumn(self, j: int, v: Vec) -> None:
        """Restore a column obtained with `getColumn()`.

        Logically collective.

        Parameters
        ----------
        j
            The index of the requested column.
        v
            The vector obtained with `getColumn()`.

        Notes
        -----
        The arguments must match the corresponding call to `getColumn()`.

        See Also
        --------
        slepc.BVRestoreColumn
        """
        ...

    def getMat(self) -> Mat:
        """Get a Mat object of dense type that shares the memory of the BV object.

        Collective.

        Returns
        -------
        Mat
            The matrix.

        Notes
        -----
        The returned matrix contains only the active columns. If the content
        of the Mat is modified, these changes are also done in the BV object.
        The user must call `restoreMat()` when no longer needed.

        See Also
        --------
        slepc.BVGetMat
        """
        ...

    def restoreMat(self, A: Mat) -> None:
        """Restore the Mat obtained with `getMat()`.

        Logically collective.

        Parameters
        ----------
        A
            The matrix obtained with `getMat()`.

        Notes
        -----
        A call to this function must match a previous call of `getMat()`.
        The effect is that the contents of the Mat are copied back to the
        BV internal data structures.

        See Also
        --------
        slepc.BVRestoreMat
        """
        ...

    def dot(self, Y: BV) -> Mat:
        """Compute the 'block-dot' product of two basis vectors objects.

        Collective.

        M = Y^H X (m_ij = y_i^H x_j) or M = Y^H B X

        Parameters
        ----------
        Y
            Left basis vectors, can be the same as self, giving
            M = X^H X.

        Returns
        -------
        Mat
            The resulting matrix.

        Notes
        -----
        This is the generalization of VecDot() for a collection of vectors,
        M = Y^H X. The result is a matrix M whose entry m_ij is equal to
        y_i^H x_j (where y_i^H denotes the conjugate transpose of y_i).

        X and Y can be the same object.

        If a non-standard inner product has been specified with setMatrix(),
        then the result is M = Y^H B X. In this case, both X and Y must have
        the same associated matrix.

        Only rows (resp. columns) of M starting from ly (resp. lx) are computed,
        where ly (resp. lx) is the number of leading columns of Y (resp. X).

        See Also
        --------
        slepc.BVDot
        """
        ...

    def matProject(self, A: Mat | None, Y: BV) -> Mat:
        """Compute the projection of a matrix onto a subspace.

        Collective.

        M = Y^H A X

        Parameters
        ----------
        A
            Matrix to be projected.
        Y
            Left basis vectors, can be the same as self, giving
            M = X^H A X.

        Returns
        -------
        Mat
            Projection of the matrix A onto the subspace.

        See Also
        --------
        slepc.BVMatProject
        """
        ...

    # --- Matrix Multiplication ---

    def matMult(self, A: Mat, Y: BV | None = None) -> BV:
        """Compute the matrix-vector product for each column, Y = A V.

        Neighbor-wise collective.

        Parameters
        ----------
        A
            The matrix.
        Y
            Optional result BV object.

        Returns
        -------
        BV
            The result.

        Notes
        -----
        Only active columns (excluding the leading ones) are processed.

        It is possible to choose whether the computation is done column by column
        or using dense matrices using the options database keys:

            -bv_matmult_vecs
            -bv_matmult_mat

        The default is bv_matmult_mat.

        See Also
        --------
        slepc.BVMatMult
        """
        ...

    def matMultHermitianTranspose(self, A: Mat, Y: BV | None = None) -> BV:
        """Pre-multiplication with the conjugate transpose of a matrix.

        Neighbor-wise collective.

        Y = A^H V.

        Parameters
        ----------
        A
            The matrix.
        Y
            Optional result BV object.

        Returns
        -------
        BV
            The result.

        Notes
        -----
        Only active columns (excluding the leading ones) are processed.

        As opposed to matMult(), this operation is always done by column by
        column, with a sequence of calls to MatMultHermitianTranspose().

        See Also
        --------
        slepc.BVMatMultHermitianTranspose
        """
        ...

    def matMultColumn(self, A: Mat, j: int) -> None:
        """Mat-vec product for a column, storing the result in the next column.

        Neighbor-wise collective.

        v_{j+1} = A v_j.

        Parameters
        ----------
        A
            The matrix.
        j
            Index of column.

        See Also
        --------
        slepc.BVMatMultColumn
        """
        ...

    def matMultTransposeColumn(self, A: Mat, j: int) -> None:
        """Transpose matrix-vector product for a specified column.

        Neighbor-wise collective.

        Store the result in the next column: v_{j+1} = A^T v_j.

        Parameters
        ----------
        A
            The matrix.
        j
            Index of column.

        See Also
        --------
        slepc.BVMatMultTransposeColumn
        """
        ...

    def matMultHermitianTransposeColumn(self, A: Mat, j: int) -> None:
        """Conjugate-transpose matrix-vector product for a specified column.

        Neighbor-wise collective.

        Store the result in the next column: v_{j+1} = A^H v_j.

        Parameters
        ----------
        A
            The matrix.
        j
            Index of column.

        See Also
        --------
        slepc.BVMatMultHermitianTransposeColumn
        """
        ...

    # --- Mult Operations ---

    def mult(self, alpha: Scalar, beta: Scalar, X: BV, Q: Mat | None) -> None:
        """Compute Y = beta Y + alpha X Q.

        Logically collective.

        Parameters
        ----------
        alpha
            Coefficient that multiplies X.
        beta
            Coefficient that multiplies Y.
        X
            Input basis vectors.
        Q
            Input matrix, if not given the identity matrix is assumed.

        See Also
        --------
        slepc.BVMult
        """
        ...

    def multInPlace(self, Q: Mat, s: int, e: int) -> None:
        """Update a set of vectors as V(:,s:e-1) = V Q(:,s:e-1).

        Logically collective.

        Parameters
        ----------
        Q
            A sequential dense matrix.
        s
            First column to be overwritten.
        e
            Last column to be overwritten.

        See Also
        --------
        slepc.BVMultInPlace
        """
        ...

    def multColumn(
        self, alpha: Scalar, beta: Scalar, j: int, q: Sequence[Scalar]
    ) -> None:
        """Compute y = beta y + alpha X q.

        Logically collective.

        Compute y = beta y + alpha X q, where y is the j-th column.

        Parameters
        ----------
        alpha
            Coefficient that multiplies X.
        beta
            Coefficient that multiplies y.
        j
            The column index.
        q
            Input coefficients.

        See Also
        --------
        slepc.BVMultColumn
        """
        ...

    def multVec(self, alpha: Scalar, beta: Scalar, y: Vec, q: Sequence[Scalar]) -> None:
        """Compute y = beta y + alpha X q.

        Logically collective.

        Parameters
        ----------
        alpha
            Coefficient that multiplies X.
        beta
            Coefficient that multiplies y.
        y
            Input/output vector.
        q
            Input coefficients.

        See Also
        --------
        slepc.BVMultVec
        """
        ...

    # --- Norms ---

    def normColumn(self, j: int, norm_type: NormTypeSpec = None) -> float:
        """Compute the vector norm of a selected column.

        Collective.

        Parameters
        ----------
        j
            Index of column.
        norm_type
            The norm type.

        Returns
        -------
        float
            The norm.

        Notes
        -----
        The norm of V_j is computed (NORM_1, NORM_2, or NORM_INFINITY).

        If a non-standard inner product has been specified with BVSetMatrix(),
        then the returned value is sqrt(V_j^H B V_j),
        where B is the inner product matrix (argument 'type' is ignored).

        See Also
        --------
        slepc.BVNormColumn
        """
        ...

    def norm(self, norm_type: NormTypeSpec = None) -> float:
        """Compute the matrix norm of the BV.

        Collective.

        Parameters
        ----------
        norm_type
            The norm type.

        Returns
        -------
        float
            The norm.

        Notes
        -----
        All active columns (except the leading ones) are considered as a
        matrix. The allowed norms are NORM_1, NORM_FROBENIUS, and
        NORM_INFINITY.

        This operation fails if a non-standard inner product has been specified
        with BVSetMatrix().

        See Also
        --------
        slepc.BVNorm
        """
        ...

    # --- Resize ---

    def resize(self, m: int, copy: bool = True) -> None:
        """Change the number of columns.

        Collective.

        Parameters
        ----------
        m
            The new number of columns.
        copy
            A flag indicating whether current values should be kept.

        Notes
        -----
        Internal storage is reallocated. If copy is True, then the contents are
        copied to the leading part of the new space.

        See Also
        --------
        slepc.BVResize
        """
        ...

    # --- Random ---

    def setRandom(self) -> None:
        """Set the active columns of the BV to random numbers.

        Logically collective.

        Notes
        -----
        All active columns (except the leading ones) are modified.

        See Also
        --------
        slepc.BVSetRandom
        """
        ...

    def setRandomNormal(self) -> None:
        """Set the active columns of the BV to normal random numbers.

        Logically collective.

        Notes
        -----
        All active columns (except the leading ones) are modified.

        See Also
        --------
        slepc.BVSetRandomNormal
        """
        ...

    def setRandomSign(self) -> None:
        """Set the entries of a BV to values 1 or -1 with equal probability.

        Logically collective.

        Notes
        -----
        All active columns (except the leading ones) are modified.

        See Also
        --------
        slepc.BVSetRandomSign
        """
        ...

    def setRandomColumn(self, j: int) -> None:
        """Set one column of the BV to random numbers.

        Logically collective.

        Parameters
        ----------
        j
            Column number to be set.

        See Also
        --------
        slepc.BVSetRandomColumn
        """
        ...

    def setRandomCond(self, condn: float) -> None:
        """Set the columns of a BV to random numbers.

        Logically collective.

        The generated matrix has a prescribed condition number.

        Parameters
        ----------
        condn
            Condition number.

        See Also
        --------
        slepc.BVSetRandomCond
        """
        ...

    def setRandomContext(self, rnd: Random) -> None:
        """Set the Random object associated with the BV.

        Collective.

        To be used in operations that need random numbers.

        Parameters
        ----------
        rnd
            The random number generator context.

        See Also
        --------
        slepc.BVSetRandomContext
        """
        ...

    def getRandomContext(self) -> Random:
        """Get the Random object associated with the BV.

        Collective.

        Returns
        -------
        Random
            The random number generator context.

        See Also
        --------
        slepc.BVGetRandomContext
        """
        ...

    # --- Orthogonalization Operations ---

    def orthogonalizeVec(self, v: Vec) -> tuple[float, bool]:
        """Orthogonalize a vector with respect to a set of vectors.

        Collective.

        Parameters
        ----------
        v
            Vector to be orthogonalized, modified on return.

        Returns
        -------
        norm : float
            The norm of the resulting vector.
        lindep : bool
            Flag indicating that refinement did not improve the
            quality of orthogonalization.

        Notes
        -----
        This function applies an orthogonal projector to project vector
        v onto the orthogonal complement of the span of the columns
        of the BV.

        This routine does not normalize the resulting vector.

        See Also
        --------
        slepc.BVOrthogonalizeVec
        """
        ...

    def orthogonalizeColumn(self, j: int) -> tuple[float, bool]:
        """Orthogonalize a column vector with respect to the previous ones.

        Collective.

        Parameters
        ----------
        j
            Index of the column to be orthogonalized.

        Returns
        -------
        norm : float
            The norm of the resulting vector.
        lindep : bool
            Flag indicating that refinement did not improve the
            quality of orthogonalization.

        Notes
        -----
        This function applies an orthogonal projector to project vector
        V_j onto the orthogonal complement of the span of the columns
        V[0..j-1], where V[.] are the vectors of the BV.
        The columns V[0..j-1] are assumed to be mutually orthonormal.

        This routine does not normalize the resulting vector.

        See Also
        --------
        slepc.BVOrthogonalizeColumn
        """
        ...

    def orthonormalizeColumn(self, j: int, replace: bool = False) -> tuple[float, bool]:
        """Orthonormalize a column vector with respect to the previous ones.

        Collective.

        This is equivalent to a call to `orthogonalizeColumn()` followed by a
        call to `scaleColumn()` with the reciprocal of the norm.

        Parameters
        ----------
        j
            Index of the column to be orthonormalized.
        replace
            Whether it is allowed to set the vector randomly.

        Returns
        -------
        norm : float
            The norm of the resulting vector.
        lindep : bool
            Flag indicating that refinement did not improve the
            quality of orthogonalization.

        See Also
        --------
        slepc.BVOrthonormalizeColumn
        """
        ...

    def orthogonalize(self, R: Mat | None = None, **kargs: Any) -> None:
        """Orthogonalize all columns (except leading ones) (QR decomposition).

        Collective.

        Parameters
        ----------
        R
            A sequential dense matrix.

        Notes
        -----
        The output satisfies V_0 = V R (where V_0 represent the
        input V) and V' V = I.

        See Also
        --------
        slepc.BVOrthogonalize
        """
        ...

    # --- Properties ---

    @property
    def sizes(self) -> tuple[tuple[int, int], int]:
        """Basis vectors local and global sizes, and the number of columns."""
        ...

    @property
    def size(self) -> tuple[int, int]:
        """Basis vectors global size."""
        ...

    @property
    def local_size(self) -> int:
        """Basis vectors local size."""
        ...

    @property
    def column_size(self) -> int:
        """Basis vectors column size."""
        ...



"""Type stubs for slepc4py.SLEPc module."""

# Main classes
from .BV import BV
from .DS import DS
from .EPS import EPS
from .FN import FN
from .LME import LME
from .MFN import MFN
from .NEP import NEP
from .PEP import PEP
from .RG import RG
from .ST import ST
from .SVD import SVD
from .Sys import Sys
from .Util import Util

# Module constants (from PETSc)
from petsc4py.PETSc.Const import (
    DECIDE,
    DEFAULT,
    DETERMINE,
    CURRENT,
)

__all__ = [
    # Main classes
    "BV",
    "DS",
    "EPS",
    "FN",
    "LME",
    "MFN",
    "NEP",
    "PEP",
    "RG",
    "ST",
    "SVD",
    "Sys",
    "Util",
    # Module constants
    "DECIDE",
    "DEFAULT",
    "DETERMINE",
    "CURRENT",
]



cdef extern from * nogil:

    ctypedef char* SlepcLMEType "const char*"
    SlepcLMEType LMEKRYLOV

    ctypedef enum SlepcLMEConvergedReason "LMEConvergedReason":
        LME_CONVERGED_TOL
        LME_DIVERGED_ITS
        LME_DIVERGED_BREAKDOWN
        LME_CONVERGED_ITERATING

    ctypedef enum SlepcLMEProblemType "LMEProblemType":
        LME_LYAPUNOV
        LME_SYLVESTER
        LME_GEN_LYAPUNOV
        LME_GEN_SYLVESTER
        LME_DT_LYAPUNOV
        LME_STEIN

    ctypedef PetscErrorCode (*SlepcLMECtxDel)(void*)
    ctypedef PetscErrorCode (*SlepcLMEMonitorFunction)(SlepcLME,
                                            PetscInt,
                                            PetscReal,
                                            void*) except PETSC_ERR_PYTHON

    PetscErrorCode LMECreate(MPI_Comm,SlepcLME*)
    PetscErrorCode LMEDestroy(SlepcLME*)
    PetscErrorCode LMEReset(SlepcLME)
    PetscErrorCode LMEView(SlepcLME,PetscViewer)

    PetscErrorCode LMESetType(SlepcLME,SlepcLMEType)
    PetscErrorCode LMEGetType(SlepcLME,SlepcLMEType*)
    PetscErrorCode LMESetProblemType(SlepcLME,SlepcLMEProblemType)
    PetscErrorCode LMEGetProblemType(SlepcLME,SlepcLMEProblemType*)

    PetscErrorCode LMESetCoefficients(SlepcLME,PetscMat,PetscMat,PetscMat,PetscMat)
    PetscErrorCode LMEGetCoefficients(SlepcLME,PetscMat*,PetscMat*,PetscMat*,PetscMat*)
    PetscErrorCode LMESetRHS(SlepcLME,PetscMat)
    PetscErrorCode LMEGetRHS(SlepcLME,PetscMat*)
    PetscErrorCode LMESetSolution(SlepcLME,PetscMat)
    PetscErrorCode LMEGetSolution(SlepcLME,PetscMat*)
    PetscErrorCode LMEGetErrorEstimate(SlepcLME,PetscReal*)
    PetscErrorCode LMEComputeError(SlepcLME,PetscReal*)

    PetscErrorCode LMESetOptionsPrefix(SlepcLME,char*)
    PetscErrorCode LMEGetOptionsPrefix(SlepcLME,char*[])
    PetscErrorCode LMESetFromOptions(SlepcLME)
    PetscErrorCode LMEAppendOptionsPrefix(SlepcLME,char*)
    PetscErrorCode LMESetUp(SlepcLME)
    PetscErrorCode LMESolve(SlepcLME)

    PetscErrorCode LMESetBV(SlepcLME,SlepcBV)
    PetscErrorCode LMEGetBV(SlepcLME,SlepcBV*)
    PetscErrorCode LMESetFN(SlepcLME,SlepcFN)
    PetscErrorCode LMEGetFN(SlepcLME,SlepcFN*)
    PetscErrorCode LMESetTolerances(SlepcLME,PetscReal,PetscInt)
    PetscErrorCode LMEGetTolerances(SlepcLME,PetscReal*,PetscInt*)
    PetscErrorCode LMESetDimensions(SlepcLME,PetscInt)
    PetscErrorCode LMEGetDimensions(SlepcLME,PetscInt*)

    PetscErrorCode LMESetErrorIfNotConverged(SlepcLME,PetscBool)
    PetscErrorCode LMEGetErrorIfNotConverged(SlepcLME,PetscBool*)

    PetscErrorCode LMEMonitorSet(SlepcLME,SlepcLMEMonitorFunction,void*,SlepcLMECtxDel)
    PetscErrorCode LMEMonitorCancel(SlepcLME)
    PetscErrorCode LMEGetIterationNumber(SlepcLME,PetscInt*)

    PetscErrorCode LMEGetConvergedReason(SlepcLME,SlepcLMEConvergedReason*)

# -----------------------------------------------------------------------------

cdef inline LME ref_LME(SlepcLME lme):
    cdef LME ob = <LME> LME()
    ob.lme = lme
    CHKERR( PetscINCREF(ob.obj) )
    return ob

# -----------------------------------------------------------------------------

cdef PetscErrorCode LME_Monitor(
    SlepcLME    lme,
    PetscInt    it,
    PetscReal   errest,
    void        *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef LME Lme = ref_LME(lme)
    cdef object monitorlist = Lme.get_attr('__monitor__')
    if monitorlist is None: return PETSC_SUCCESS
    for (monitor, args, kargs) in monitorlist:
        monitor(Lme, toInt(it), toReal(errest), *args, **kargs)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------



cdef extern from * nogil:

    ctypedef char* SlepcBVType "const char*"
    SlepcBVType BVMAT
    SlepcBVType BVSVEC
    SlepcBVType BVVECS
    SlepcBVType BVCONTIGUOUS
    SlepcBVType BVTENSOR

    ctypedef enum SlepcBVOrthogType "BVOrthogType":
        BV_ORTHOG_CGS
        BV_ORTHOG_MGS

    ctypedef enum SlepcBVOrthogRefineType "BVOrthogRefineType":
        BV_ORTHOG_REFINE_IFNEEDED
        BV_ORTHOG_REFINE_NEVER
        BV_ORTHOG_REFINE_ALWAYS

    ctypedef enum SlepcBVOrthogBlockType "BVOrthogBlockType":
        BV_ORTHOG_BLOCK_GS
        BV_ORTHOG_BLOCK_CHOL
        BV_ORTHOG_BLOCK_TSQR
        BV_ORTHOG_BLOCK_TSQRCHOL
        BV_ORTHOG_BLOCK_SVQB

    ctypedef enum SlepcBVMatMultType "BVMatMultType":
        BV_MATMULT_VECS
        BV_MATMULT_MAT

    ctypedef enum SlepcBVSVDMethod "BVSVDMethod":
        BV_SVD_METHOD_REFINE
        BV_SVD_METHOD_QR
        BV_SVD_METHOD_QR_CAA

    PetscErrorCode BVCreate(MPI_Comm,SlepcBV*)
    PetscErrorCode BVCreateMat(SlepcBV,PetscMat*)
    PetscErrorCode BVDuplicate(SlepcBV,SlepcBV*)
    PetscErrorCode BVDuplicateResize(SlepcBV,PetscInt,SlepcBV*)
    PetscErrorCode BVCopy(SlepcBV,SlepcBV)
    PetscErrorCode BVView(SlepcBV,PetscViewer)
    PetscErrorCode BVDestroy(SlepcBV*)
    PetscErrorCode BVSetType(SlepcBV,SlepcBVType)
    PetscErrorCode BVGetType(SlepcBV,SlepcBVType*)
    PetscErrorCode BVSetSizes(SlepcBV,PetscInt,PetscInt,PetscInt)
    PetscErrorCode BVSetSizesFromVec(SlepcBV,PetscVec,PetscInt)
    PetscErrorCode BVGetSizes(SlepcBV,PetscInt*,PetscInt*,PetscInt*)
    PetscErrorCode BVResize(SlepcBV,PetscInt,PetscBool)
    PetscErrorCode BVSetLeadingDimension(SlepcBV,PetscInt)
    PetscErrorCode BVGetLeadingDimension(SlepcBV,PetscInt*)

    PetscErrorCode BVSetOptionsPrefix(SlepcBV,char[])
    PetscErrorCode BVGetOptionsPrefix(SlepcBV,char*[])
    PetscErrorCode BVAppendOptionsPrefix(SlepcBV,char[])
    PetscErrorCode BVSetFromOptions(SlepcBV)

    PetscErrorCode BVSetOrthogonalization(SlepcBV,SlepcBVOrthogType,SlepcBVOrthogRefineType,PetscReal,SlepcBVOrthogBlockType)
    PetscErrorCode BVGetOrthogonalization(SlepcBV,SlepcBVOrthogType*,SlepcBVOrthogRefineType*,PetscReal*,SlepcBVOrthogBlockType*)
    PetscErrorCode BVSetMatMultMethod(SlepcBV,SlepcBVMatMultType)
    PetscErrorCode BVGetMatMultMethod(SlepcBV,SlepcBVMatMultType*)

    PetscErrorCode BVSetRandom(SlepcBV)
    PetscErrorCode BVSetRandomNormal(SlepcBV)
    PetscErrorCode BVSetRandomSign(SlepcBV)
    PetscErrorCode BVSetRandomColumn(SlepcBV,PetscInt)
    PetscErrorCode BVSetRandomCond(SlepcBV,PetscReal)
    PetscErrorCode BVSetRandomContext(SlepcBV,PetscRandom)
    PetscErrorCode BVGetRandomContext(SlepcBV,PetscRandom*)

    PetscErrorCode BVSetMatrix(SlepcBV,PetscMat,PetscBool)
    PetscErrorCode BVGetMatrix(SlepcBV,PetscMat*,PetscBool*)
    PetscErrorCode BVApplyMatrix(SlepcBV,PetscVec,PetscVec)

    PetscErrorCode BVSetActiveColumns(SlepcBV,PetscInt,PetscInt)
    PetscErrorCode BVGetActiveColumns(SlepcBV,PetscInt*,PetscInt*)
    PetscErrorCode BVSetDefiniteTolerance(SlepcBV,PetscReal)
    PetscErrorCode BVGetDefiniteTolerance(SlepcBV,PetscReal*)

    PetscErrorCode BVCreateVec(SlepcBV,PetscVec*)
    PetscErrorCode BVSetVecType(SlepcBV,PetscVecType)
    PetscErrorCode BVGetVecType(SlepcBV,PetscVecType*)
    PetscErrorCode BVInsertVec(SlepcBV,PetscInt,PetscVec)
    PetscErrorCode BVInsertVecs(SlepcBV,PetscInt,PetscInt*,PetscVec*,PetscBool)
    PetscErrorCode BVInsertConstraints(SlepcBV,PetscInt*,PetscVec*)
    PetscErrorCode BVSetNumConstraints(SlepcBV,PetscInt)
    PetscErrorCode BVGetNumConstraints(SlepcBV,PetscInt*)
    PetscErrorCode BVGetColumn(SlepcBV,PetscInt,PetscVec*)
    PetscErrorCode BVRestoreColumn(SlepcBV,PetscInt,PetscVec*)
    PetscErrorCode BVCopyVec(SlepcBV,PetscInt,PetscVec)
    PetscErrorCode BVCopyColumn(SlepcBV,PetscInt,PetscInt)

    PetscErrorCode BVDot(SlepcBV,SlepcBV,PetscMat)
    PetscErrorCode BVDotVec(SlepcBV,PetscVec,PetscScalar*)
    PetscErrorCode BVDotColumn(SlepcBV,PetscInt,PetscScalar*)

    PetscErrorCode BVMatProject(SlepcBV,PetscMat,SlepcBV,PetscMat)
    PetscErrorCode BVMatMult(SlepcBV,PetscMat,SlepcBV)
    PetscErrorCode BVMatMultHermitianTranspose(SlepcBV,PetscMat,SlepcBV)
    PetscErrorCode BVMatMultColumn(SlepcBV,PetscMat,PetscInt)
    PetscErrorCode BVMatMultTransposeColumn(SlepcBV,PetscMat,PetscInt)
    PetscErrorCode BVMatMultHermitianTransposeColumn(SlepcBV,PetscMat,PetscInt)

    PetscErrorCode BVMult(SlepcBV,PetscScalar,PetscScalar,SlepcBV,PetscMat)
    PetscErrorCode BVMultColumn(SlepcBV,PetscScalar,PetscScalar,PetscInt,PetscScalar*)
    PetscErrorCode BVMultInPlace(SlepcBV,PetscMat,PetscInt,PetscInt)
    PetscErrorCode BVMultVec(SlepcBV,PetscScalar,PetscScalar,PetscVec,PetscScalar*)

    PetscErrorCode BVScaleColumn(SlepcBV,PetscInt,PetscScalar)
    PetscErrorCode BVScale(SlepcBV,PetscScalar)

    PetscErrorCode BVNormColumn(SlepcBV,PetscInt,PetscNormType,PetscReal*)
    PetscErrorCode BVNorm(SlepcBV,PetscNormType,PetscReal*)

    PetscErrorCode BVOrthogonalizeVec(SlepcBV,PetscVec,PetscScalar*,PetscReal*,PetscBool*)
    PetscErrorCode BVOrthogonalizeColumn(SlepcBV,PetscInt,PetscScalar*,PetscReal*,PetscBool*)
    PetscErrorCode BVOrthonormalizeColumn(SlepcBV,PetscInt,PetscBool,PetscReal*,PetscBool*)
    PetscErrorCode BVOrthogonalize(SlepcBV,PetscMat)

    PetscErrorCode BVCreateFromMat(PetscMat,SlepcBV*)
    PetscErrorCode BVGetMat(SlepcBV,PetscMat*)
    PetscErrorCode BVRestoreMat(SlepcBV,PetscMat*)

cdef inline PetscErrorCode BV_Sizes(
    object size,
    PetscInt *_n,
    PetscInt *_N,
    ) except PETSC_ERR_PYTHON:
    # unpack and get local and global sizes
    cdef PetscInt n=PETSC_DECIDE, N=PETSC_DECIDE
    cdef object on, oN
    try:
        on, oN = size
    except (TypeError, ValueError):
        on = None; oN = size
    if on is not None: n = asInt(on)
    if oN is not None: N = asInt(oN)
    if n==PETSC_DECIDE and N==PETSC_DECIDE: raise ValueError(
        "local and global sizes cannot be both 'DECIDE'")
    # return result to the caller
    if _n != NULL: _n[0] = n
    if _N != NULL: _N[0] = N
    return PETSC_SUCCESS

# --------------------------------------------------------------------

# unary operations

cdef BV bv_pos(BV self):
    cdef BV bv = type(self)()
    CHKERR(BVDuplicate(self.bv, &bv.bv))
    CHKERR(BVCopy(self.bv, bv.bv))
    return bv

cdef BV bv_neg(BV self):
    cdef BV bv = <BV> bv_pos(self)
    CHKERR(BVScale(bv.bv, -1))
    return bv

# inplace binary operations

cdef BV bv_iadd(BV self, other):
    cdef PetscScalar alpha = 1, beta = 1
    cdef BV bv
    if isinstance(other, BV):
        alpha = 1; bv = other
        CHKERR(BVMult(self.bv, alpha, beta, bv.bv, NULL))
        return self
    elif isinstance(other, (tuple, list)):
        other, bv = other
        alpha = asScalar(other)
        CHKERR(BVMult(self.bv, alpha, beta, bv.bv, NULL))
        return self
    return NotImplemented

cdef BV bv_isub(BV self, other):
    cdef PetscScalar alpha = 1, beta = 1
    cdef BV bv
    if isinstance(other, BV):
        alpha = 1; bv = other
        CHKERR(BVMult(self.bv, -alpha, beta, bv.bv, NULL))
        return self
    elif isinstance(other, (tuple, list)):
        other, bv = other
        alpha = asScalar(other)
        CHKERR(BVMult(self.bv, -alpha, beta, bv.bv, NULL))
        return self
    return NotImplemented

cdef BV bv_imul(BV self, other):
    self.scale(other)
    return self

cdef BV bv_idiv(BV self, other):
    other = 1/other
    self.scale(other)
    return self

# binary operations

cdef BV bv_add(BV self, other):
    return bv_iadd(bv_pos(self), other)

cdef BV bv_sub(BV self, other):
    return bv_isub(bv_pos(self), other)

cdef BV bv_mul(BV self, other):
    return bv_imul(bv_pos(self), other)

cdef BV bv_div(BV self, other):
    return bv_idiv(bv_pos(self), other)

# reflected binary operations

cdef BV bv_radd(BV self, other):
    return bv_add(self, other)

cdef BV bv_rsub(BV self, other):
    cdef BV bv = <BV> bv_sub(self, other)
    CHKERR(BVScale(bv.bv, -1))
    return bv

cdef BV bv_rmul(BV self, other):
    return bv_mul(self, other)

cdef BV bv_rdiv(BV self, other):
    <void>self; <void>other # unused
    return NotImplemented

